import datetime
import os
import time

import numpy as np
import termcolor
import yaml

current_logger = None


class Logger:
    """Logger used to display and save logs, and save experiment configs."""

    def __init__(self, path=None, width=60, script_path=None, config=None):
        self.path = path or str(time.time())
        self.log_file_path = os.path.join(self.path, "log.csv")

        # Save the launch script.
        if script_path:
            with open(script_path, "r") as script_file:
                script = script_file.read()
                try:
                    os.makedirs(self.path, exist_ok=True)
                except Exception:
                    pass
                script_path = os.path.join(self.path, "script.py")
                with open(script_path, "w") as config_file:
                    config_file.write(script)
                log(f"Script file saved to {script_path}")

        # Save the configuration.
        if config:
            try:
                os.makedirs(self.path, exist_ok=True)
            except Exception:
                pass
            config_path = os.path.join(self.path, "config.yaml")
            with open(config_path, "w") as config_file:
                yaml.dump(config, config_file)
            log(f"Config file saved to {config_path}")

        self.known_keys = set()
        self.stat_keys = set()
        self.epoch_dict = {}
        self.width = width
        self.last_epoch_progress = None
        self.start_time = time.time()

    def store(self, key, value, stats=False):
        """Keeps named values during an epoch."""

        if key not in self.epoch_dict:
            self.epoch_dict[key] = [value]
            if stats:
                self.stat_keys.add(key)
        else:
            self.epoch_dict[key].append(value)

    def dump(self):
        """Displays and saves the values at the end of an epoch."""

        # Compute statistics if needed.
        keys = list(self.epoch_dict.keys())
        for key in keys:
            values = self.epoch_dict[key]
            if key in self.stat_keys:
                self.epoch_dict[key + "/mean"] = np.mean(values)
                self.epoch_dict[key + "/std"] = np.std(values)
                self.epoch_dict[key + "/min"] = np.min(values)
                self.epoch_dict[key + "/max"] = np.max(values)
                self.epoch_dict[key + "/size"] = len(values)
                del self.epoch_dict[key]
            else:
                self.epoch_dict[key] = np.mean(values)

        # Check if new keys were added.
        new_keys = [
            key for key in self.epoch_dict.keys() if key not in self.known_keys
        ]
        if new_keys:
            first_row = len(self.known_keys) == 0
            if not first_row:
                print()
                warning(f"Logging new keys {new_keys}")
            # List the keys and prepare the display layout.
            for key in new_keys:
                self.known_keys.add(key)
            self.final_keys = list(sorted(self.known_keys))
            self.console_formats = []
            known_keys = set()
            for key in self.final_keys:
                *left_keys, right_key = key.split("/")
                for i, k in enumerate(left_keys):
                    left_key = "/".join(left_keys[: i + 1])
                    if left_key not in known_keys:
                        left = "  " * i + k.replace("_", " ")
                        self.console_formats.append((left, None))
                        known_keys.add(left_key)
                indent = "  " * len(left_keys)
                right_key = right_key.replace("_", " ")
                self.console_formats.append((indent + right_key, key))

        # Display the values following the layout.
        print()
        for left, key in self.console_formats:
            if key:
                val = self.epoch_dict.get(key)
                str_type = str(type(val))
                if "tensorflow" in str_type:
                    warning(f"Logging TensorFlow tensor {key}")
                elif "torch" in str_type:
                    warning(f"Logging Torch tensor {key}")
                if np.issubdtype(type(val), np.floating):
                    right = f"{val:8.3g}"
                elif np.issubdtype(type(val), np.integer):
                    right = f"{val:,}"
                else:
                    right = str(val)
                spaces = " " * (self.width - len(left) - len(right))
                print(left + spaces + right)
            else:
                spaces = " " * (self.width - len(left))
                print(left + spaces)
        print()

        # Save the data to the log file
        vals = [self.epoch_dict.get(key) for key in self.final_keys]
        if new_keys:
            if first_row:
                log(f"Logging data to {self.log_file_path}")
                try:
                    os.makedirs(self.path, exist_ok=True)
                except Exception:
                    pass
                with open(self.log_file_path, "w") as file:
                    file.write(",".join(self.final_keys) + "\n")
                    file.write(",".join(map(str, vals)) + "\n")
            else:
                with open(self.log_file_path, "r") as file:
                    lines = file.read().splitlines()
                old_keys = lines[0].split(",")
                old_lines = [line.split(",") for line in lines[1:]]
                new_indices = []
                j = 0
                for i, key in enumerate(self.final_keys):
                    if key == old_keys[j]:
                        j += 1
                    else:
                        new_indices.append(i)
                assert j == len(old_keys)
                for line in old_lines:
                    for i in new_indices:
                        line.insert(i, "None")
                with open(self.log_file_path, "w") as file:
                    file.write(",".join(self.final_keys) + "\n")
                    for line in old_lines:
                        file.write(",".join(line) + "\n")
                    file.write(",".join(map(str, vals)) + "\n")
        else:
            with open(self.log_file_path, "a") as file:
                file.write(",".join(map(str, vals)) + "\n")

        self.epoch_dict.clear()
        self.last_epoch_progress = None
        self.last_epoch_time = time.time()

    def show_progress(
        self,
        steps,
        num_epoch_steps,
        num_steps,
        color="white",
        on_color="on_blue",
    ):
        """Shows a progress bar for the current epoch and total training."""

        epoch_steps = (steps - 1) % num_epoch_steps + 1
        epoch_progress = int(self.width * epoch_steps / num_epoch_steps)
        if epoch_progress != self.last_epoch_progress:
            current_time = time.time()
            seconds = current_time - self.start_time
            seconds_per_step = seconds / steps
            epoch_rem_steps = num_epoch_steps - epoch_steps
            epoch_rem_secs = max(epoch_rem_steps * seconds_per_step, 0)
            epoch_rem_secs = datetime.timedelta(seconds=epoch_rem_secs + 1e-6)
            epoch_rem_secs = str(epoch_rem_secs)[:-7]
            total_rem_steps = num_steps - steps
            total_rem_secs = max(total_rem_steps * seconds_per_step, 0)
            total_rem_secs = datetime.timedelta(seconds=total_rem_secs)
            total_rem_secs = str(total_rem_secs)[:-7]
            msg = f"Time left:  epoch {epoch_rem_secs}  total {total_rem_secs}"
            msg = msg.center(self.width)
            print(
                termcolor.colored(
                    "\r" + msg[:epoch_progress], color, on_color
                ),
                end="",
            )
            print(msg[epoch_progress:], sep="", end="")
            self.last_epoch_progress = epoch_progress


def initialize(*args, **kwargs):
    global current_logger
    current_logger = Logger(*args, **kwargs)
    return current_logger


def get_current_logger():
    global current_logger
    if current_logger is None:
        current_logger = Logger()
    return current_logger


def store(*args, **kwargs):
    logger = get_current_logger()
    return logger.store(*args, **kwargs)


def dump(*args, **kwargs):
    logger = get_current_logger()
    return logger.dump(*args, **kwargs)


def show_progress(*args, **kwargs):
    logger = get_current_logger()
    return logger.show_progress(*args, **kwargs)


def get_path():
    logger = get_current_logger()
    return logger.path


def log(msg, color="green"):
    print(termcolor.colored(msg, color, attrs=["bold"]))


def warning(msg, color="yellow"):
    print(termcolor.colored("Warning: " + msg, color, attrs=["bold"]))


def error(msg, color="red"):
    print(termcolor.colored("Error: " + msg, color, attrs=["bold"]))



import numpy as np

from deprl.vendor.tonic.replays.utils import flatten_batch, lambda_returns


class Segment:
    """Replay storing recent transitions for on-policy learning."""

    def __init__(
        self,
        size=4096,
        batch_iterations=80,
        batch_size=None,
        discount_factor=0.99,
        trace_decay=0.97,
    ):
        self.max_size = size
        self.batch_iterations = batch_iterations
        self.batch_size = batch_size
        self.discount_factor = discount_factor
        self.trace_decay = trace_decay

    def initialize(self, seed=None):
        self.np_random = np.random.RandomState(seed)
        self.buffers = None
        self.index = 0

    def ready(self):
        return self.index == self.max_size

    def store(self, **kwargs):
        if self.buffers is None:
            self.num_workers = len(list(kwargs.values())[0])
            self.buffers = {}
            for key, val in kwargs.items():
                shape = (self.max_size,) + np.array(val).shape
                self.buffers[key] = np.zeros(shape, np.float32)
        for key, val in kwargs.items():
            self.buffers[key][self.index] = val
        self.index += 1

    def get_full(self, *keys):
        self.index = 0

        if "advantages" in keys:
            advs = self.buffers["returns"] - self.buffers["values"]
            std = advs.std()
            if std != 0:
                advs = (advs - advs.mean()) / std
            self.buffers["advantages"] = advs

        return {k: flatten_batch(self.buffers[k]) for k in keys}

    def get(self, *keys):
        """Get mini-batches from named buffers."""

        batch = self.get_full(*keys)

        if self.batch_size is None:
            for _ in range(self.batch_iterations):
                yield batch
        else:
            size = self.max_size * self.num_workers
            all_indices = np.arange(size)
            for _ in range(self.batch_iterations):
                self.np_random.shuffle(all_indices)
                for i in range(0, size, self.batch_size):
                    indices = all_indices[i : i + self.batch_size]
                    yield {k: v[indices] for k, v in batch.items()}

    def compute_returns(self, values, next_values):
        shape = self.buffers["rewards"].shape
        self.buffers["values"] = values.reshape(shape)
        self.buffers["next_values"] = next_values.reshape(shape)
        self.buffers["returns"] = lambda_returns(
            values=self.buffers["values"],
            next_values=self.buffers["next_values"],
            rewards=self.buffers["rewards"],
            resets=self.buffers["resets"],
            terminations=self.buffers["terminations"],
            discount_factor=self.discount_factor,
            trace_decay=self.trace_decay,
        )



import numpy as np
from matplotlib import pyplot as plt


class ScoreKeeper:
    """
    Plotting utility. Works best with dictionaries of information that
    should be stored and plotted.
    """

    def __init__(self):
        self.reset()
        self.steps = 0

    def store(self, data_dict):
        if not self.data_dict:
            self._store_initial_dict(data_dict)
        self._append_score(data_dict)

    def get_score(self, name, mean=True):
        try:
            if mean:
                return np.mean(self.data_dict[name])
            else:
                return np.array(self.data_dict[name])
        except KeyError:
            raise Exception("Invalid key specified")

    def keys(self):
        return self.data_dict.keys()

    def reset(self):
        self.data_dict = None
        self.fig = None
        self.axs = None

    def print_all(self):
        print("Printing rwd_dict now:")
        for k in self.keys():
            print(f"For {k=} we have score={self.get_score(k)}")

    def plot(self, names):
        if type(names) is not list:
            names = [names]
        fig, axs = plt.subplots(len(names), 1)
        for name in names:
            if len(names) > 1:
                axs[names.index(name)].plot(self.get_score(name, mean=False))
                axs[names.index(name)].set_title(name)
            else:
                axs.plot(self.get_score(name, mean=False))
                axs.set_title(name)
        plt.show()

    def adapt_plot(self, names):
        if not self.steps % 10:
            data = [self.get_score(name, mean=False) for name in names]
            x = np.arange(0, data[0].shape[0])
            if self.fig is None:
                plt.ion()
                self.fig, self.axs = plt.subplots(len(names), 1)
                self.lines = [
                    ax.plot(x, d)[0] for ax, d in zip(self.axs, data)
                ]
                # self.axs[0].set_ylim([-500, 0])
                # self.axs[1].set_ylim([-0.5, 0.2])
                # self.axs[2].set_ylim([0, 1.1])
                # self.axs[4].set_ylim([-0.002, 0.001])
                for name, ax in zip(names, self.axs):
                    ax.set_title(name)
                    ax.set_xlim([0, 1000])
            else:
                for line, d in zip(self.lines, data):
                    line.set_ydata(d)
                    line.set_xdata(x)
            self.fig.canvas.draw()
            self.fig.canvas.flush_events()
            for ax in self.axs:
                ax.relim()
                ax.autoscale_view()
            if len(data[0]) > 1000:
                self._clear()
        self.steps += 1
        # time.sleep(0.001)

    def _store_initial_dict(self, data_dict):
        self.data_dict = {k: [] for k in data_dict.keys()}

    def _append_score(self, data_dict):
        for k, v in data_dict.items():
            if k not in self.data_dict.keys():
                self.data_dict[k] = []
            self.data_dict[k].append(v)

    def _clear(self):
        self.data_dict = None



"""Environment wrappers."""

import gym
import numpy as np


class ActionRescaler(gym.ActionWrapper):
    """Rescales actions from [-1, 1]^n to the true action space.
    The baseline agents return actions in [-1, 1]^n."""

    def __init__(self, env):
        assert isinstance(env.action_space, gym.spaces.Box)
        super().__init__(env)
        high = np.ones(env.action_space.shape, dtype=np.float32)
        self.action_space = gym.spaces.Box(low=-high, high=high)
        true_low = env.action_space.low
        true_high = env.action_space.high
        self.bias = (true_high + true_low) / 2
        self.scale = (true_high - true_low) / 2

    def action(self, action):
        return self.bias + self.scale * np.clip(action, -1, 1)


class TimeFeature(gym.Wrapper):
    """Adds a notion of time in the observations.
    It can be used in terminal timeout settings to get Markovian MDPs.
    """

    def __init__(self, env, max_steps, low=-1, high=1):
        super().__init__(env)
        dtype = self.observation_space.dtype
        self.observation_space = gym.spaces.Box(
            low=np.append(self.observation_space.low, low).astype(dtype),
            high=np.append(self.observation_space.high, high).astype(dtype),
        )
        self.max_episode_steps = max_steps
        self.steps = 0
        self.low = low
        self.high = high

    def reset(self, **kwargs):
        self.steps = 0
        observation = self.env.reset(**kwargs)
        observation = np.append(observation, self.low)
        return observation

    def step(self, action):
        assert self.steps < self.max_episode_steps
        observation, reward, done, info = self.env.step(action)
        self.steps += 1
        prop = self.steps / self.max_episode_steps
        v = self.low + (self.high - self.low) * prop
        observation = np.append(observation, v)
        return observation, reward, done, info



import tensorflow as tf
from tonic.tensorflow import updaters


class VRegression:
    def __init__(self, loss=None, optimizer=None, gradient_clip=0):
        self.loss = loss or tf.keras.losses.MeanSquaredError()
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.critic.trainable_variables

    @tf.function
    def __call__(self, observations, returns):
        with tf.GradientTape() as tape:
            values = self.model.critic(observations)
            loss = self.loss(returns, values)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss, v=values)


class QRegression:
    def __init__(self, loss=None, optimizer=None, gradient_clip=0):
        self.loss = loss or tf.keras.losses.MeanSquaredError()
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.critic.trainable_variables

    @tf.function
    def __call__(self, observations, actions, returns):
        with tf.GradientTape() as tape:
            values = self.model.critic(observations, actions)
            loss = self.loss(returns, values)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss, q=values)


class DeterministicQLearning:
    def __init__(self, loss=None, optimizer=None, gradient_clip=0):
        self.loss = loss or tf.keras.losses.MeanSquaredError()
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.critic.trainable_variables

    @tf.function
    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        next_actions = self.model.target_actor(next_observations)
        next_values = self.model.target_critic(next_observations, next_actions)
        returns = rewards + discounts * next_values

        with tf.GradientTape() as tape:
            values = self.model.critic(observations, actions)
            loss = self.loss(returns, values)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss, q=values)


class DistributionalDeterministicQLearning:
    def __init__(self, optimizer=None, gradient_clip=0):
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.critic.trainable_variables

    @tf.function
    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        next_actions = self.model.target_actor(next_observations)
        next_value_distributions = self.model.target_critic(
            next_observations, next_actions
        )

        values = next_value_distributions.values
        returns = rewards[:, None] + discounts[:, None] * values
        targets = next_value_distributions.project(returns)

        with tf.GradientTape() as tape:
            value_distributions = self.model.critic(observations, actions)
            losses = tf.nn.softmax_cross_entropy_with_logits(
                logits=value_distributions.logits, labels=targets
            )
            loss = tf.reduce_mean(losses)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss)


class TargetActionNoise:
    def __init__(self, scale=0.2, clip=0.5):
        self.scale = scale
        self.clip = clip

    def __call__(self, actions):
        noises = self.scale * tf.random.normal(actions.shape)
        noises = tf.clip_by_value(noises, -self.clip, self.clip)
        actions = actions + noises
        return tf.clip_by_value(actions, -1, 1)


class TwinCriticDeterministicQLearning:
    def __init__(
        self,
        loss=None,
        optimizer=None,
        target_action_noise=None,
        gradient_clip=0,
    ):
        self.loss = loss or tf.keras.losses.MeanSquaredError()
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.target_action_noise = target_action_noise or TargetActionNoise(
            scale=0.2, clip=0.5
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        variables_1 = self.model.critic_1.trainable_variables
        variables_2 = self.model.critic_2.trainable_variables
        self.variables = variables_1 + variables_2

    @tf.function
    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        next_actions = self.model.target_actor(next_observations)
        next_actions = self.target_action_noise(next_actions)
        next_values_1 = self.model.target_critic_1(
            next_observations, next_actions
        )
        next_values_2 = self.model.target_critic_2(
            next_observations, next_actions
        )
        next_values = tf.minimum(next_values_1, next_values_2)
        returns = rewards + discounts * next_values

        with tf.GradientTape() as tape:
            values_1 = self.model.critic_1(observations, actions)
            values_2 = self.model.critic_2(observations, actions)
            loss_1 = self.loss(returns, values_1)
            loss_2 = self.loss(returns, values_2)
            loss = loss_1 + loss_2

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss, q1=values_1, q2=values_2)


class TwinCriticSoftQLearning:
    def __init__(
        self, loss=None, optimizer=None, entropy_coeff=0.2, gradient_clip=0
    ):
        self.loss = loss or tf.keras.losses.MeanSquaredError()
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.entropy_coeff = entropy_coeff
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        variables_1 = self.model.critic_1.trainable_variables
        variables_2 = self.model.critic_2.trainable_variables
        self.variables = variables_1 + variables_2

    @tf.function
    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        next_distributions = self.model.actor(next_observations)
        if hasattr(next_distributions, "sample_with_log_prob"):
            outs = next_distributions.sample_with_log_prob()
            next_actions, next_log_probs = outs
        else:
            next_actions = next_distributions.sample()
            next_log_probs = next_distributions.log_prob(next_actions)
        next_values_1 = self.model.target_critic_1(
            next_observations, next_actions
        )
        next_values_2 = self.model.target_critic_2(
            next_observations, next_actions
        )
        next_values = tf.minimum(next_values_1, next_values_2)
        returns = rewards + discounts * (
            next_values - self.entropy_coeff * next_log_probs
        )

        with tf.GradientTape() as tape:
            values_1 = self.model.critic_1(observations, actions)
            values_2 = self.model.critic_2(observations, actions)
            loss_1 = self.loss(returns, values_1)
            loss_2 = self.loss(returns, values_2)
            loss = loss_1 + loss_2

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss, q1=values_1, q2=values_2)


class ExpectedSARSA:
    def __init__(
        self, num_samples=20, loss=None, optimizer=None, gradient_clip=0
    ):
        self.num_samples = num_samples
        self.loss = loss or tf.keras.losses.MeanSquaredError()
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.critic.trainable_variables

    @tf.function
    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        # Approximate the expected next values.
        next_target_distributions = self.model.target_actor(next_observations)
        next_actions = next_target_distributions.sample(self.num_samples)
        next_actions = updaters.merge_first_two_dims(next_actions)
        next_observations = updaters.tile(next_observations, self.num_samples)
        next_observations = updaters.merge_first_two_dims(next_observations)
        next_values = self.model.target_critic(next_observations, next_actions)
        next_values = tf.reshape(next_values, (self.num_samples, -1))
        next_values = tf.reduce_mean(next_values, axis=0)
        returns = rewards + discounts * next_values

        with tf.GradientTape() as tape:
            values = self.model.critic(observations, actions)
            loss = self.loss(returns, values)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss, q=values)


class TwinCriticDistributionalDeterministicQLearning:
    def __init__(
        self, optimizer=None, target_action_noise=None, gradient_clip=0
    ):
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.target_action_noise = target_action_noise or TargetActionNoise(
            scale=0.2, clip=0.5
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        variables_1 = self.model.critic_1.trainable_variables
        variables_2 = self.model.critic_2.trainable_variables
        self.variables = variables_1 + variables_2

    @tf.function
    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        next_actions = self.model.target_actor(next_observations)
        next_actions = self.target_action_noise(next_actions)
        next_value_distributions_1 = self.model.target_critic_1(
            next_observations, next_actions
        )
        next_value_distributions_2 = self.model.target_critic_2(
            next_observations, next_actions
        )

        values = next_value_distributions_1.values
        returns = rewards[:, None] + discounts[:, None] * values
        targets_1 = next_value_distributions_1.project(returns)
        targets_2 = next_value_distributions_2.project(returns)
        next_values_1 = next_value_distributions_1.mean()
        next_values_2 = next_value_distributions_2.mean()
        twin_next_values = tf.concat(
            [next_values_1[None], next_values_2[None]], axis=0
        )
        indices = tf.argmin(twin_next_values, axis=0, output_type=tf.int32)
        twin_targets = tf.concat([targets_1[None], targets_2[None]], axis=0)
        batch_size = tf.shape(observations)[0]
        indices = tf.stack([indices, tf.range(batch_size)], axis=-1)
        targets = tf.gather_nd(twin_targets, indices)

        with tf.GradientTape() as tape:
            value_distributions_1 = self.model.critic_1(observations, actions)
            losses_1 = tf.nn.softmax_cross_entropy_with_logits(
                logits=value_distributions_1.logits, labels=targets
            )
            value_distributions_2 = self.model.critic_2(observations, actions)
            losses_2 = tf.nn.softmax_cross_entropy_with_logits(
                logits=value_distributions_2.logits, labels=targets
            )
            loss = tf.reduce_mean(losses_1) + tf.reduce_mean(losses_2)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(
            loss=loss,
            q1=value_distributions_1.mean(),
            q2=value_distributions_2.mean(),
        )



import numpy as np

from deprl.vendor.tonic import logger


def test_mujoco(env, agent, steps, params=None, test_episodes=10):
    """
    Tests the agent on the test environment.
    """
    # Start the environment.
    if not hasattr(env, "test_observations"):
        # Dont use dep in evaluation
        env.test_observations, _ = env.start()
        assert len(env.test_observations) == 1

    # Test loop.
    for _ in range(test_episodes):
        metrics = {
            "test/episode_score": 0,
            "test/episode_length": 0,
            "test/effort": 0,
            "test/terminated": 0,
        }
        while True:
            # Select an action.
            actions = agent.test_step(env.test_observations, steps)
            assert not np.isnan(actions.sum())
            logger.store("test/action", actions, stats=True)

            # Take a step in the environment.
            env.test_observations, _, info = env.step(actions)

            # Update metrics
            metrics["test/episode_score"] += info["rewards"][0]
            metrics["test/episode_length"] += 1
            if env.environments[0].sim.model.na > 0:
                metrics["test/effort"] += np.mean(
                    np.square(env.environments[0].unwrapped.sim.data.act)
                )
            metrics["test/terminated"] += int(info["terminations"])

            if info["resets"][0]:
                break
        # Log the data.Average over episode length here
        metrics["test/terminated"] /= metrics["test/episode_length"]
        metrics["test/effort"] /= metrics["test/episode_length"]
        # average over episodes in logger
        for k, v in metrics.items():
            logger.store(k, v, stats=True)
    return metrics


def test_dm_control(env, agent, steps, params=None, test_episodes=10):
    """
    Tests the agent on the test environment.
    """
    # Start the environment.
    if not hasattr(env, "test_observations"):
        # Dont use dep in evaluation
        env.test_observations, _ = env.start()
        assert len(env.test_observations) == 1

    max_reward = 0
    # Test loop.
    for _ in range(test_episodes):
        metrics = {
            "test/episode_score": 0,
            "test/episode_length": 0,
            "test/effort": 0,
            "test/terminated": 0,
        }
        while True:
            # Select an action.
            actions = agent.test_step(env.test_observations, steps)
            assert not np.isnan(actions.sum())
            logger.store("test/action", actions, stats=True)

            # Take a step in the environment.
            env.test_observations, _, info = env.step(actions)

            # Update metrics
            metrics["test/episode_score"] += info["rewards"][0]
            metrics["test/episode_length"] += 1
            metrics["test/effort"] += np.mean(
                np.square(env.environments[0].muscle_activity())
            )
            metrics["test/terminated"] += int(info["terminations"])
            max_reward = max(max_reward, info["rewards"][0])

            if info["resets"][0]:
                break
        # Log the data.Average over episode length here
        metrics["test/terminated"] /= metrics["test/episode_length"]
        metrics["test/effort"] /= metrics["test/episode_length"]
        # average over episodes in logger
        for k, v in metrics.items():
            logger.store(k, v, stats=True)
    # max over all episodes
    logger.store("test/max_reward", max_reward, stats=False)
    return metrics



import tensorflow as tf
from tonic.tensorflow import models


class ValueHead(tf.keras.Model):
    def __init__(self, dense_kwargs=None):
        super().__init__()
        if dense_kwargs is None:
            dense_kwargs = models.default_dense_kwargs()
        self.v_layer = tf.keras.layers.Dense(1, **dense_kwargs)

    def initialize(self, return_normalizer=None):
        self.return_normalizer = return_normalizer

    def call(self, inputs):
        out = self.v_layer(inputs)
        out = tf.squeeze(out, -1)
        if self.return_normalizer:
            out = self.return_normalizer(out)
        return out


class CategoricalWithSupport:
    def __init__(self, values, logits):
        self.values = values
        self.logits = logits
        self.probabilities = tf.nn.softmax(logits)

    def mean(self):
        return tf.reduce_sum(self.probabilities * self.values, axis=-1)

    def project(self, returns):
        vmin, vmax = self.values[0], self.values[-1]
        d_pos = tf.concat([self.values, vmin[None]], 0)[1:]
        d_pos = (d_pos - self.values)[None, :, None]
        d_neg = tf.concat([vmax[None], self.values], 0)[:-1]
        d_neg = (self.values - d_neg)[None, :, None]

        clipped_returns = tf.clip_by_value(returns, vmin, vmax)
        delta_values = clipped_returns[:, None] - self.values[None, :, None]
        delta_sign = tf.cast(delta_values >= 0, tf.float32)
        delta_hat = (delta_sign * delta_values / d_pos) - (
            (1 - delta_sign) * delta_values / d_neg
        )
        delta_clipped = tf.clip_by_value(1 - delta_hat, 0, 1)

        return tf.reduce_sum(delta_clipped * self.probabilities[:, None], 2)


class DistributionalValueHead(tf.keras.Model):
    def __init__(self, vmin, vmax, num_atoms, dense_kwargs=None):
        super().__init__()
        if dense_kwargs is None:
            dense_kwargs = models.default_dense_kwargs()
        self.distributional_layer = tf.keras.layers.Dense(
            num_atoms, **dense_kwargs
        )
        self.values = tf.cast(tf.linspace(vmin, vmax, num_atoms), tf.float32)

    def initialize(self, return_normalizer=None):
        if return_normalizer:
            raise ValueError(
                "Return normalizers cannot be used with distributional value"
                "heads."
            )

    def call(self, inputs):
        logits = self.distributional_layer(inputs)
        return CategoricalWithSupport(values=self.values, logits=logits)


class Critic(tf.keras.Model):
    def __init__(self, encoder, torso, head):
        super().__init__()
        self.encoder = encoder
        self.torso = torso
        self.head = head

    def initialize(
        self,
        observation_space,
        action_space,
        observation_normalizer=None,
        return_normalizer=None,
    ):
        self.encoder.initialize(observation_normalizer)
        self.head.initialize(return_normalizer)

    def call(self, *inputs):
        out = self.encoder(*inputs)
        out = self.torso(out)
        return self.head(out)



import copy

import torch

from deprl.vendor.tonic.torch import models


class ActorCritic(torch.nn.Module):
    def __init__(
        self,
        actor,
        critic,
        observation_normalizer=None,
        return_normalizer=None,
    ):
        super().__init__()
        self.actor = actor
        self.critic = critic
        self.observation_normalizer = observation_normalizer
        self.return_normalizer = return_normalizer

    def initialize(self, observation_space, action_space):
        if self.observation_normalizer:
            self.observation_normalizer.initialize(observation_space.shape)
        self.actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.critic.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )


class ActorCriticWithTargets(torch.nn.Module):
    def __init__(
        self,
        actor,
        critic,
        observation_normalizer=None,
        return_normalizer=None,
        target_coeff=0.005,
    ):
        super().__init__()
        self.actor = actor
        self.critic = critic
        self.target_actor = copy.deepcopy(actor)
        self.target_critic = copy.deepcopy(critic)
        self.observation_normalizer = observation_normalizer
        self.return_normalizer = return_normalizer
        self.target_coeff = target_coeff

    def initialize(self, observation_space, action_space):
        if self.observation_normalizer:
            self.observation_normalizer.initialize(observation_space.shape)
        self.actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.critic.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.target_actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.target_critic.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.online_variables = models.trainable_variables(self.actor)
        self.online_variables += models.trainable_variables(self.critic)
        self.target_variables = models.trainable_variables(self.target_actor)
        self.target_variables += models.trainable_variables(self.target_critic)
        for target in self.target_variables:
            target.requires_grad = False
        self.assign_targets()

    def assign_targets(self):
        for o, t in zip(self.online_variables, self.target_variables):
            t.data.copy_(o.data)

    def update_targets(self):
        with torch.no_grad():
            for o, t in zip(self.online_variables, self.target_variables):
                t.data.mul_(1 - self.target_coeff)
                t.data.add_(self.target_coeff * o.data)


class ActorTwinCriticWithTargets(torch.nn.Module):
    def __init__(
        self,
        actor,
        critic,
        observation_normalizer=None,
        return_normalizer=None,
        target_coeff=0.005,
    ):
        super().__init__()
        self.actor = actor
        self.critic_1 = critic
        self.critic_2 = copy.deepcopy(critic)
        self.target_actor = copy.deepcopy(actor)
        self.target_critic_1 = copy.deepcopy(critic)
        self.target_critic_2 = copy.deepcopy(critic)
        self.observation_normalizer = observation_normalizer
        self.return_normalizer = return_normalizer
        self.target_coeff = target_coeff

    def initialize(self, observation_space, action_space):
        if self.observation_normalizer:
            self.observation_normalizer.initialize(observation_space.shape)
        self.actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.critic_1.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.critic_2.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.target_actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.target_critic_1.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.target_critic_2.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.online_variables = models.trainable_variables(self.actor)
        self.online_variables += models.trainable_variables(self.critic_1)
        self.online_variables += models.trainable_variables(self.critic_2)
        self.target_variables = models.trainable_variables(self.target_actor)
        self.target_variables += models.trainable_variables(
            self.target_critic_1
        )
        self.target_variables += models.trainable_variables(
            self.target_critic_2
        )
        for target in self.target_variables:
            target.requires_grad = False
        self.assign_targets()

    def assign_targets(self):
        for o, t in zip(self.online_variables, self.target_variables):
            t.data.copy_(o.data)

    def update_targets(self):
        with torch.no_grad():
            for o, t in zip(self.online_variables, self.target_variables):
                t.data.mul_(1 - self.target_coeff)
                t.data.add_(self.target_coeff * o.data)



import csv
import os


def check_if_csv_has_updated(csv_path, last_update):
    """
    Checks if the csv file has been updated since the last update.
    """
    current_update = os.path.getmtime(csv_path)
    if current_update > last_update:
        return current_update, True
    return current_update, False


def load_csv(csv_path):
    """
    Loads the csv file and returns the data.
    """
    with open(csv_path, "r") as f:
        reader = csv.reader(f)
        data = list(reader)
    return data


def load_csv_to_dict(csv_path):
    """
    Converts the csv file to a dictionary.
    """
    data = load_csv(csv_path)
    keys = data[0]
    data = data[1:]
    data = [list(map(float, x)) for x in data if x[0] != "None"]
    data = {k: [x[idx] for x in data] for idx, k in enumerate(keys)}
    return data



import torch

from deprl.vendor.tonic.torch import models, normalizers, updaters
from deprl.vendor.tonic.utils import logger

from .ddpg import DDPG


def default_model():
    return models.ActorTwinCriticWithTargets(
        actor=models.Actor(
            encoder=models.ObservationEncoder(),
            torso=models.MLP((256, 256), torch.nn.ReLU),
            head=models.DeterministicPolicyHead(),
        ),
        critic=models.Critic(
            encoder=models.ObservationActionEncoder(),
            torso=models.MLP((256, 256), torch.nn.ReLU),
            head=models.ValueHead(),
        ),
        observation_normalizer=normalizers.MeanStd(),
    )


class TD3(DDPG):
    """Twin Delayed Deep Deterministic Policy Gradient.
    TD3: https://arxiv.org/pdf/1802.09477.pdf
    """

    def __init__(
        self,
        model=None,
        replay=None,
        exploration=None,
        actor_updater=None,
        critic_updater=None,
        delay_steps=2,
    ):
        model = model or default_model()
        critic_updater = (
            critic_updater or updaters.TwinCriticDeterministicQLearning()
        )
        super().__init__(
            model=model,
            replay=replay,
            exploration=exploration,
            actor_updater=actor_updater,
            critic_updater=critic_updater,
        )
        self.delay_steps = delay_steps
        self.model.critic = self.model.critic_1

    def _update(self, steps):
        keys = (
            "observations",
            "actions",
            "next_observations",
            "rewards",
            "discounts",
        )
        for i, batch in enumerate(self.replay.get(*keys, steps=steps)):
            batch = {k: torch.as_tensor(v) for k, v in batch.items()}
            if (i + 1) % self.delay_steps == 0:
                infos = self._update_actor_critic(**batch)
            else:
                infos = dict(critic=self.critic_updater(**batch))
            for key in infos:
                for k, v in infos[key].items():
                    logger.store(key + "/" + k, v.numpy())

        # Update the normalizers.
        if self.model.observation_normalizer:
            self.model.observation_normalizer.update()
        if self.model.return_normalizer:
            self.model.return_normalizer.update()



import tensorflow as tf


class ObservationEncoder(tf.keras.Model):
    def initialize(self, observation_normalizer=None):
        self.observation_normalizer = observation_normalizer

    def call(self, observations):
        if self.observation_normalizer:
            observations = self.observation_normalizer(observations)
        return observations


class ObservationActionEncoder(tf.keras.Model):
    def initialize(self, observation_normalizer=None):
        self.observation_normalizer = observation_normalizer

    def call(self, observations, actions):
        if self.observation_normalizer:
            observations = self.observation_normalizer(observations)
        return tf.concat([observations, actions], axis=-1)



"""Environment builders for popular domains."""

import os
from types import SimpleNamespace

import gym.wrappers
import numpy as np

from deprl.vendor.tonic import environments
from deprl.vendor.tonic.utils import logger


def gym_environment(*args, **kwargs):
    """Returns a wrapped Gym environment."""

    def _builder(*args, **kwargs):
        return gym.make(*args, **kwargs)

    return build_environment(_builder, *args, **kwargs)


def bullet_environment(*args, **kwargs):
    """Returns a wrapped PyBullet environment."""

    def _builder(*args, **kwargs):
        import pybullet_envs  # noqa

        return gym.make(*args, **kwargs)

    return build_environment(_builder, *args, **kwargs)


def control_suite_environment(*args, **kwargs):
    """Returns a wrapped Control Suite environment."""

    def _builder(name, *args, **kwargs):
        domain, task = name.split("-")
        environment = ControlSuiteEnvironment(
            domain_name=domain, task_name=task, *args, **kwargs
        )
        time_limit = int(environment.environment._step_limit)
        environment.spec = SimpleNamespace(max_episode_steps=time_limit)
        return gym.wrappers.TimeLimit(environment, time_limit)

    return build_environment(_builder, *args, **kwargs)


def build_environment(
    builder,
    name,
    terminal_timeouts=False,
    time_feature=False,
    max_episode_steps="default",
    scaled_actions=True,
    *args,
    **kwargs,
):
    """Builds and wrap an environment.
    Time limits can be properly handled with terminal_timeouts=False or
    time_feature=True, see https://arxiv.org/pdf/1712.00378.pdf for more
    details.
    """

    # Build the environment.
    environment = builder(name, *args, **kwargs)

    # Get the default time limit.
    if max_episode_steps == "default":
        max_episode_steps = environment._max_episode_steps

    # Remove the TimeLimit wrapper if needed.
    if not terminal_timeouts:
        assert type(environment) == gym.wrappers.TimeLimit, environment
        environment = environment.env

    # Add time as a feature if needed.
    if time_feature:
        environment = environments.wrappers.TimeFeature(
            environment, max_episode_steps
        )

    # Scale actions from [-1, 1]^n to the true action space if needed.
    if scaled_actions:
        environment = environments.wrappers.ActionRescaler(environment)

    environment.name = name
    environment.max_episode_steps = max_episode_steps

    return environment


def _flatten_observation(observation):
    """Turns OrderedDict observations into vectors."""
    observation = [
        np.array([o]) if np.isscalar(o) else o.ravel()
        for o in observation.values()
    ]
    return np.concatenate(observation, axis=0)


class ControlSuiteEnvironment(gym.core.Env):
    """Turns a Control Suite environment into a Gym environment."""

    def __init__(
        self,
        domain_name,
        task_name,
        task_kwargs=None,
        visualize_reward=True,
        environment_kwargs=None,
    ):
        from dm_control import suite
        from dm_control.rl.control import PhysicsError

        self.environment = suite.load(
            domain_name=domain_name,
            task_name=task_name,
            task_kwargs=task_kwargs,
            visualize_reward=visualize_reward,
            environment_kwargs=environment_kwargs,
        )

        # Create the observation space.
        observation_spec = self.environment.observation_spec()
        dim = sum(
            [int(np.prod(spec.shape)) for spec in observation_spec.values()]
        )
        high = np.full(dim, np.inf, np.float32)
        self.observation_space = gym.spaces.Box(-high, high, dtype=np.float32)

        # Create the action space.
        action_spec = self.environment.action_spec()
        self.action_space = gym.spaces.Box(
            action_spec.minimum, action_spec.maximum, dtype=np.float32
        )
        self.error = PhysicsError

    def seed(self, seed):
        self.environment.task._random = np.random.RandomState(seed)

    def step(self, action):
        try:
            time_step = self.environment.step(action)
            observation = _flatten_observation(time_step.observation)
            reward = time_step.reward

            # Remove terminations from timeouts.
            done = time_step.last()
            if done:
                done = self.environment.task.get_termination(
                    self.environment.physics
                )
                done = done is not None

            self.last_time_step = time_step

        # In case MuJoCo crashed.
        except self.error as e:
            path = logger.get_path()
            os.makedirs(path, exist_ok=True)
            save_path = os.path.join(path, "crashes.txt")
            error = str(e)
            with open(save_path, "a") as file:
                file.write(error + "\n")
            logger.error(error)
            observation = _flatten_observation(self.last_time_step.observation)
            observation = np.zeros_like(observation)
            reward = 0.0
            done = True

        return observation, reward, done, {}

    def reset(self):
        time_step = self.environment.reset()
        self.last_time_step = time_step
        return _flatten_observation(time_step.observation)

    def render(self, mode="rgb_array", height=None, width=None, camera_id=0):
        """Returns RGB frames from a camera."""
        assert mode == "rgb_array"
        return self.environment.physics.render(
            height=height, width=width, camera_id=camera_id
        )


# Aliases.
Gym = gym_environment
Bullet = bullet_environment
ControlSuite = control_suite_environment



import numpy as np
import tensorflow as tf


class MeanStd(tf.keras.Model):
    def __init__(self, mean=0, std=1, clip=None, shape=None):
        super().__init__(name="global_mean_std")
        self.mean = mean
        self.std = std
        self.clip = clip
        self.count = 0
        self.new_sum = 0
        self.new_sum_sq = 0
        self.new_count = 0
        self.eps = 1e-2
        if shape:
            self.initialize(shape)

    def initialize(self, shape):
        if isinstance(self.mean, (int, float)):
            self.mean = np.full(shape, self.mean, np.float32)
        else:
            self.mean = np.array(self.mean, np.float32)
        if isinstance(self.std, (int, float)):
            self.std = np.full(shape, self.std, np.float32)
        else:
            self.std = np.array(self.std, np.float32)
        self.mean_sq = np.square(self.mean)
        self._mean = tf.Variable(self.mean, trainable=False, name="mean")
        self._std = tf.Variable(self.std, trainable=False, name="std")

    def call(self, val):
        val = (val - self._mean) / self._std
        if self.clip is not None:
            val = tf.clip_by_value(val, -self.clip, self.clip)
        return val

    def unnormalize(self, val):
        return val * self._std + self._mean

    def record(self, values):
        for val in values:
            self.new_sum += val
            self.new_sum_sq += np.square(val)
            self.new_count += 1

    # Careful: do not use in @tf.function
    def update(self):
        new_count = self.count + self.new_count
        new_mean = self.new_sum / self.new_count
        new_mean_sq = self.new_sum_sq / self.new_count
        w_old = self.count / new_count
        w_new = self.new_count / new_count
        self.mean = w_old * self.mean + w_new * new_mean
        self.mean_sq = w_old * self.mean_sq + w_new * new_mean_sq
        self.std = self._compute_std(self.mean, self.mean_sq)
        self.count = new_count
        self.new_count = 0
        self.new_sum = 0
        self.new_sum_sq = 0
        self._update(self.mean.astype(np.float32), self.std.astype(np.float32))

    def _compute_std(self, mean, mean_sq):
        var = mean_sq - np.square(mean)
        var = np.maximum(var, 0)
        std = np.sqrt(var)
        std = np.maximum(std, self.eps)
        return std

    @tf.function
    def _update(self, mean, std):
        self._mean.assign(mean)
        self._std.assign(std)



"""Script used to train agents."""
import argparse
import os
import traceback

import torch
import yaml

from deprl import custom_distributed
from deprl.utils import prepare_params
from deprl.vendor import tonic


def maybe_load_checkpoint(
    header,
    agent,
    environment,
    trainer,
    time_dict,
    checkpoint_path,
    checkpoint,
    eff_path,
):
    if os.path.isdir(checkpoint_path):
        tonic.logger.log(f"Loading experiment from {eff_path}")
        # Use no checkpoint, the agent is freshly created.
        if checkpoint == "none":
            tonic.logger.log("Not loading any weights")

        else:
            # List all the checkpoints.
            checkpoint_ids = []
            for file in os.listdir(checkpoint_path):
                if file[:5] == "step_":
                    checkpoint_id = file.split(".")[0]
                    checkpoint_ids.append(int(checkpoint_id[5:]))

            if checkpoint_ids:
                # Use the last checkpoint.
                if checkpoint == "last":
                    checkpoint_id = max(checkpoint_ids)
                    checkpoint_path = os.path.join(
                        checkpoint_path, f"step_{checkpoint_id}"
                    )

                # Use the specified checkpoint.
                else:
                    checkpoint_id = int(checkpoint)
                    if checkpoint_id in checkpoint_ids:
                        checkpoint_path = os.path.join(
                            checkpoint_path, f"step_{checkpoint_id}"
                        )
                    else:
                        tonic.logger.error(
                            f"Checkpoint {checkpoint_id} not found in {checkpoint_path}"
                        )
                        checkpoint_path = None
            else:
                tonic.logger.error(f"No checkpoint found in {checkpoint_path}")
                checkpoint_path = None

        # Load the experiment configuration.
        arguments_path = os.path.join(eff_path, "config.yaml")
        with open(arguments_path, "r") as config_file:
            config = yaml.load(config_file, Loader=yaml.FullLoader)
        config = argparse.Namespace(**config)

        header = header or config.header
        agent = agent or config.agent
        environment = environment or config.test_environment
        environment = environment or config.environment
        trainer = trainer or config.trainer
        return header, agent, environment, trainer, time_dict, checkpoint_path

    else:
        checkpoint_path = None
        return header, agent, environment, trainer, time_dict, checkpoint_path


def train(
    orig_params,
    header,
    agent,
    environment,
    test_environment,
    trainer,
    before_training,
    after_training,
    parallel,
    sequential,
    seed,
    name,
    environment_name,
    checkpoint,
    path,
    preid=0,
    env_args=None,
):
    """Trains an agent on an environment."""
    # Capture the arguments to save them, e.g. to play with the trained agent.
    # TODO fix this mess and do it properly
    args = dict(locals())
    del args["orig_params"]
    if args["env_args"]:
        args["env_args"] = dict(args["env_args"])
        if "target" in args["env_args"]:
            args["env_args"]["target"] = list(args["env_args"]["target"])
        if "rew_args" in args["env_args"]:
            args["env_args"]["rew_args"] = dict(args["env_args"]["rew_args"])
    if "mpo_args" in orig_params:
        args["mpo_args"] = dict(orig_params["mpo_args"])

    eff_path = os.path.join(path, environment_name, name)
    # Process the checkpoint path same way as in tonic.play
    checkpoint_path = os.path.join(eff_path, "checkpoints")
    time_dict = {"steps": 0, "epochs": 0, "episodes": 0}
    (
        header,
        agent,
        environment,
        trainer,
        time_dict,
        checkpoint_path,
    ) = maybe_load_checkpoint(
        header,
        agent,
        environment,
        trainer,
        time_dict,
        checkpoint_path,
        checkpoint,
        eff_path,
    )
    # Run the header first, e.g. to load an ML framework.
    if header:
        exec(header)
    if "env_args" not in locals():
        env_args = {}
    else:
        env_args = dict(env_args)
    env_args["starting_steps"] = time_dict["steps"]
    # Build the training environment.
    _environment = environment
    environment = custom_distributed.distribute(
        dict(
            env=_environment,
            preid=preid,
            parallel=parallel,
            sequential=sequential,
        ),
        parallel,
        sequential,
        env_args=env_args,
    )
    environment.initialize(seed=seed)
    # Build the testing environment.
    _test_environment = test_environment if test_environment else _environment
    test_env_args = dict(env_args)
    if "Baoding" in _test_environment:
        test_env_args["step_fn"] = "step_baoding_test"
    else:
        test_env_args["step_fn"] = "step_die"
    test_environment = custom_distributed.distribute(
        dict(env=_test_environment, preid=preid + 1000000),
        env_args=test_env_args,
    )
    test_environment.initialize(seed=seed + 1000000)

    # Build the agent.
    if not agent:
        raise ValueError("No agent specified.")
    agent = eval(agent)
    if "mpo_args" in orig_params:
        agent.set_params(**orig_params["mpo_args"])
    agent.initialize(
        observation_space=environment.observation_space,
        action_space=environment.action_space,
        seed=seed,
    )
    if hasattr(agent, "expl") and "DEP" in orig_params:
        agent.expl.set_params(orig_params["DEP"])
    # Load the weights of the agent form a checkpoint.
    if checkpoint_path:
        agent.load(checkpoint_path)

    # Initialize the logger to save data to the path environment/name/seed.
    if not environment_name:
        if hasattr(test_environment, "name"):
            environment_name = test_environment.name
        else:
            environment_name = test_environment.__class__.__name__
    if not name:
        if hasattr(agent, "name"):
            name = agent.name
        else:
            name = agent.__class__.__name__
        if parallel != 1 or sequential != 1:
            name += f"-{parallel}x{sequential}"
    eff_path = os.path.join(path, environment_name, name)

    tonic.logger.initialize(eff_path, script_path=__file__, config=args)

    # Build the trainer.
    trainer = trainer or "tonic.Trainer()"
    trainer = eval(trainer)
    trainer.initialize(
        agent=agent, environment=environment, test_environment=test_environment
    )

    # Run some code before training.
    if before_training:
        exec(before_training)

    # Train.
    try:
        scores = trainer.run(orig_params, **time_dict)
    except Exception as e:
        tonic.logger.log(f"trainer failed. Exception: {e}")
        traceback.print_tb(e.__traceback__)

    # Run some code after training.
    if after_training:
        exec(after_training)


if __name__ == "__main__":
    try:
        torch.zeros((0, 1), device="cuda")
        torch.set_default_tensor_type("torch.cuda.FloatTensor")
    except Exception as e:
        print(f"No cuda detected, running on cpu: {e}")
    orig_params, params = prepare_params()
    train_params = dict(orig_params["tonic"])
    train_params["path"] = orig_params["working_dir"]
    train_params["preid"] = orig_params["id"]
    if "env_args" in orig_params or "env_args" in train_params:
        train_params["env_args"] = (
            orig_params["env_args"]
            if "env_args" in orig_params
            else train_params["env_args"]
        )
    train(orig_params, **train_params)



import os
import time

import numpy as np
import torch

from deprl.custom_test_environment import test_dm_control, test_mujoco
from deprl.vendor.tonic import logger

if "ROBOHIVE_VERBOSITY" not in os.environ:
    os.environ["ROBOHIVE_VERBOSITY"] = "ALWAYS"


class Trainer:
    """Trainer used to train and evaluate an agent on an environment."""

    def __init__(
        self,
        steps=1e7,
        epoch_steps=2e4,
        save_steps=5e5,
        test_episodes=20,
        show_progress=True,
        replace_checkpoint=False,
    ):
        self.max_steps = int(steps)
        self.epoch_steps = int(epoch_steps)
        self.save_steps = int(save_steps)
        self.test_episodes = test_episodes
        self.show_progress = show_progress
        self.replace_checkpoint = replace_checkpoint

    def initialize(self, agent, environment, test_environment=None):
        self.agent = agent
        self.environment = environment
        self.test_environment = test_environment

    def run(self, params, steps=0, epochs=0, episodes=0):
        """Runs the main training loop."""

        start_time = last_epoch_time = time.time()

        # Start the environments.
        observations, tendon_states = self.environment.start()

        num_workers = len(observations)
        scores = np.zeros(num_workers)
        lengths = np.zeros(num_workers, int)
        self.steps, epoch_steps = steps, 0
        steps_since_save = 0

        while True:
            # Select actions.
            if hasattr(self.agent, "expl"):
                greedy_episode = (
                    not episodes % self.agent.expl.test_episode_every
                )
            else:
                greedy_episode = None
            assert not np.isnan(observations.sum())
            actions = self.agent.step(
                observations, self.steps, tendon_states, greedy_episode
            )
            assert not np.isnan(actions.sum())
            # raise Exception(f'{type(self.environment.environments[0])}')
            logger.store("train/action", actions, stats=True)

            # Take a step in the environments.
            observations, tendon_states, info = self.environment.step(actions)
            if "env_infos" in info:
                info.pop("env_infos")

            self.agent.update(**info, steps=self.steps)

            scores += info["rewards"]
            lengths += 1
            self.steps += num_workers
            epoch_steps += num_workers
            steps_since_save += num_workers

            # Show the progress bar.
            if self.show_progress:
                logger.show_progress(
                    self.steps, self.epoch_steps, self.max_steps
                )

            # Check the finished episodes.
            for i in range(num_workers):
                if info["resets"][i]:
                    logger.store("train/episode_score", scores[i], stats=True)
                    logger.store(
                        "train/episode_length", lengths[i], stats=True
                    )
                    scores[i] = 0
                    lengths[i] = 0
                    episodes += 1

            # End of the epoch.
            if epoch_steps >= self.epoch_steps:
                # Evaluate the agent on the test environment.
                if self.test_environment:
                    if "Control" not in str(
                        type(self.test_environment.environments[0].unwrapped)
                    ):
                        _ = test_mujoco(
                            self.test_environment, self.agent, steps, params
                        )
                    else:
                        _ = test_dm_control(
                            self.test_environment, self.agent, steps, params
                        )

                # Log the data.
                epochs += 1
                current_time = time.time()
                epoch_time = current_time - last_epoch_time
                sps = epoch_steps / epoch_time
                logger.store("train/episodes", episodes)
                logger.store("train/epochs", epochs)
                logger.store("train/seconds", current_time - start_time)
                logger.store("train/epoch_seconds", epoch_time)
                logger.store("train/epoch_steps", epoch_steps)
                logger.store("train/steps", self.steps)
                logger.store("train/worker_steps", self.steps // num_workers)
                logger.store("train/steps_per_second", sps)
                last_epoch_time = time.time()
                epoch_steps = 0

                logger.dump()

            # End of training.
            stop_training = self.steps >= self.max_steps

            # Save a checkpoint.
            if stop_training or steps_since_save >= self.save_steps:
                path = os.path.join(logger.get_path(), "checkpoints")
                if os.path.isdir(path) and self.replace_checkpoint:
                    for file in os.listdir(path):
                        if file.startswith("step_"):
                            os.remove(os.path.join(path, file))
                checkpoint_name = f"step_{self.steps}"
                save_path = os.path.join(path, checkpoint_name)
                self.agent.save(save_path)
                # logger.save(save_path)
                # self.save_time(save_path, epochs, episodes)
                steps_since_save = self.steps % self.save_steps
                current_time = time.time()

            if stop_training:
                return scores

    def save_time(self, path, epochs, episodes):
        time_path = self.get_path(path, "time")
        time_dict = {
            "epochs": epochs,
            "episodes": episodes,
            "steps": self.steps,
        }
        torch.save(time_dict, time_path)

    def get_path(self, path, post_fix):
        return path.split("step")[0] + post_fix + ".pt"



"""Some basic non-learning agents used for example for debugging."""

import numpy as np

from deprl.vendor.tonic.agents.agent import Agent


class NormalRandom(Agent):
    """Random agent producing actions from normal distributions."""

    def __init__(self, loc=0, scale=1):
        self.loc = loc
        self.scale = scale

    def initialize(self, observation_space, action_space, seed=None):
        self.action_size = action_space.shape[0]
        self.np_random = np.random.RandomState(seed)

    def step(self, observations, steps):
        return self._policy(observations)

    def test_step(self, observations, steps):
        return self._policy(observations)

    def _policy(self, observations):
        batch_size = len(observations)
        shape = (batch_size, self.action_size)
        return self.np_random.normal(self.loc, self.scale, shape)


class UniformRandom(Agent):
    """Random agent producing actions from uniform distributions."""

    def initialize(self, observation_space, action_space, seed=None):
        self.action_size = action_space.shape[0]
        self.np_random = np.random.RandomState(seed)

    def step(self, observations, steps):
        return self._policy(observations)

    def test_step(self, observations, steps):
        return self._policy(observations)

    def _policy(self, observations):
        batch_size = len(observations)
        shape = (batch_size, self.action_size)
        return self.np_random.uniform(-1, 1, shape)


class OrnsteinUhlenbeck(Agent):
    """Random agent producing correlated actions from an OU process."""

    def __init__(self, scale=0.2, clip=2, theta=0.15, dt=1e-2):
        self.scale = scale
        self.clip = clip
        self.theta = theta
        self.dt = dt

    def initialize(self, observation_space, action_space, seed=None):
        self.action_size = action_space.shape[0]
        self.np_random = np.random.RandomState(seed)
        self.train_actions = None
        self.test_actions = None

    def step(self, observations, steps):
        return self._train_policy(observations)

    def test_step(self, observations, steps):
        return self._test_policy(observations)

    def _train_policy(self, observations):
        if self.train_actions is None:
            shape = (len(observations), self.action_size)
            self.train_actions = np.zeros(shape)
        self.train_actions = self._next_actions(self.train_actions)
        return self.train_actions

    def _test_policy(self, observations):
        if self.test_actions is None:
            shape = (len(observations), self.action_size)
            self.test_actions = np.zeros(shape)
        self.test_actions = self._next_actions(self.test_actions)
        return self.test_actions

    def _next_actions(self, actions):
        noises = self.np_random.normal(size=actions.shape)
        noises = np.clip(noises, -self.clip, self.clip)
        next_actions = (1 - self.theta * self.dt) * actions
        next_actions += self.scale * np.sqrt(self.dt) * noises
        next_actions = np.clip(next_actions, -1, 1)
        return next_actions

    def update(self, observations, rewards, resets, terminations, steps):
        self.train_actions *= (1.0 - resets)[:, None]

    def test_update(self, observations, rewards, resets, terminations, steps):
        self.test_actions *= (1.0 - resets)[:, None]


class Constant(Agent):
    """Agent producing a unique constant action."""

    def __init__(self, constant=0):
        self.constant = constant

    def initialize(self, observation_space, action_space, seed=None):
        self.action_size = action_space.shape[0]

    def step(self, observations, steps):
        return self._policy(observations)

    def test_step(self, observations, steps):
        return self._policy(observations)

    def _policy(self, observations):
        shape = (len(observations), self.action_size)
        return np.full(shape, self.constant)



import tensorflow as tf
import tensorflow_probability as tfp
from tonic.tensorflow import models

FLOAT_EPSILON = 1e-8


class SquashedMultivariateNormalDiag:
    def __init__(self, loc, scale):
        self._distribution = tfp.distributions.MultivariateNormalDiag(
            loc, scale
        )

    def sample_with_log_prob(self, shape=()):
        samples = self._distribution.sample(shape)
        squashed_samples = tf.tanh(samples)
        log_probs = self._distribution.log_prob(samples)
        log_probs -= tf.reduce_sum(
            tf.math.log(1 - squashed_samples**2 + 1e-6), axis=-1
        )
        return squashed_samples, log_probs

    def sample(self, shape=()):
        samples = self._distribution.sample(shape)
        return tf.tanh(samples)

    def log_prob(self, samples):
        """Required unsquashed samples cannot be accurately recovered."""
        raise NotImplementedError(
            "Not implemented to avoid approximation errors. "
            "Use sample_with_log_prob directly."
        )

    def mode(self):
        return tf.tanh(self._distribution.mode())


class DetachedScaleGaussianPolicyHead(tf.keras.Model):
    def __init__(
        self,
        loc_activation="tanh",
        dense_loc_kwargs=None,
        log_scale_init=0.0,
        scale_min=1e-4,
        scale_max=1.0,
        distribution=tfp.distributions.MultivariateNormalDiag,
    ):
        super().__init__()
        self.loc_activation = loc_activation
        if dense_loc_kwargs is None:
            dense_loc_kwargs = models.default_dense_kwargs()
        self.dense_loc_kwargs = dense_loc_kwargs
        self.log_scale_init = log_scale_init
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.distribution = distribution

    def initialize(self, action_size):
        self.loc_layer = tf.keras.layers.Dense(
            action_size, self.loc_activation, **self.dense_loc_kwargs
        )
        log_scale = [[self.log_scale_init] * action_size]
        self.log_scale = tf.Variable(log_scale, dtype=tf.float32)

    def call(self, inputs):
        loc = self.loc_layer(inputs)
        batch_size = tf.shape(inputs)[0]
        scale = tf.math.softplus(self.log_scale) + FLOAT_EPSILON
        scale = tf.clip_by_value(scale, self.scale_min, self.scale_max)
        scale = tf.tile(scale, (batch_size, 1))
        return self.distribution(loc, scale)


class GaussianPolicyHead(tf.keras.Model):
    def __init__(
        self,
        loc_activation="tanh",
        dense_loc_kwargs=None,
        scale_activation="softplus",
        scale_min=1e-4,
        scale_max=1,
        dense_scale_kwargs=None,
        distribution=tfp.distributions.MultivariateNormalDiag,
    ):
        super().__init__()
        self.loc_activation = loc_activation
        if dense_loc_kwargs is None:
            dense_loc_kwargs = models.default_dense_kwargs()
        self.dense_loc_kwargs = dense_loc_kwargs
        self.scale_activation = scale_activation
        self.scale_min = scale_min
        self.scale_max = scale_max
        if dense_scale_kwargs is None:
            dense_scale_kwargs = models.default_dense_kwargs()
        self.dense_scale_kwargs = dense_scale_kwargs
        self.distribution = distribution

    def initialize(self, action_size):
        self.loc_layer = tf.keras.layers.Dense(
            action_size, self.loc_activation, **self.dense_loc_kwargs
        )
        self.scale_layer = tf.keras.layers.Dense(
            action_size, self.scale_activation, **self.dense_scale_kwargs
        )

    def call(self, inputs):
        loc = self.loc_layer(inputs)
        scale = self.scale_layer(inputs)
        scale = tf.clip_by_value(scale, self.scale_min, self.scale_max)
        return self.distribution(loc, scale)


class DeterministicPolicyHead(tf.keras.Model):
    def __init__(self, activation="tanh", dense_kwargs=None):
        super().__init__()
        self.activation = activation
        if dense_kwargs is None:
            dense_kwargs = models.default_dense_kwargs()
        self.dense_kwargs = dense_kwargs

    def initialize(self, action_size):
        self.action_layer = tf.keras.layers.Dense(
            action_size, self.activation, **self.dense_kwargs
        )

    def call(self, inputs):
        return self.action_layer(inputs)


class Actor(tf.keras.Model):
    def __init__(self, encoder, torso, head):
        super().__init__()
        self.encoder = encoder
        self.torso = torso
        self.head = head

    def initialize(
        self, observation_space, action_space, observation_normalizer=None
    ):
        self.encoder.initialize(observation_normalizer)
        self.head.initialize(action_space.shape[0])

    def call(self, *inputs):
        out = self.encoder(*inputs)
        out = self.torso(out)
        return self.head(out)



"""Builders for distributed training."""

import multiprocessing

import numpy as np


class Sequential:
    """A group of environments used in sequence."""

    def __init__(self, environment_builder, max_episode_steps, workers):
        self.environments = [environment_builder() for _ in range(workers)]
        self.max_episode_steps = max_episode_steps
        self.observation_space = self.environments[0].observation_space
        self.action_space = self.environments[0].action_space
        self.name = self.environments[0].name

    def initialize(self, seed):
        for i, environment in enumerate(self.environments):
            environment.seed(seed + i)

    def start(self):
        """Used once to get the initial observations."""
        observations = [env.reset() for env in self.environments]
        self.lengths = np.zeros(len(self.environments), int)
        return np.array(observations, np.float32)

    def step(self, actions):
        next_observations = []  # Observations for the transitions.
        rewards = []
        resets = []
        terminations = []
        observations = []  # Observations for the actions selection.

        for i in range(len(self.environments)):
            ob, rew, term, _ = self.environments[i].step(actions[i])

            self.lengths[i] += 1
            # Timeouts trigger resets but are not true terminations.
            reset = term or self.lengths[i] == self.max_episode_steps
            next_observations.append(ob)
            rewards.append(rew)
            resets.append(reset)
            terminations.append(term)

            if reset:
                ob = self.environments[i].reset()
                self.lengths[i] = 0

            observations.append(ob)

        observations = np.array(observations, np.float32)
        infos = dict(
            observations=np.array(next_observations, np.float32),
            rewards=np.array(rewards, np.float32),
            resets=np.array(resets, bool),
            terminations=np.array(terminations, bool),
        )
        return observations, infos

    def render(self, mode="human", *args, **kwargs):
        outs = []
        for env in self.environments:
            out = env.render(mode=mode, *args, **kwargs)
            outs.append(out)
        if mode != "human":
            return np.array(outs)


class Parallel:
    """A group of sequential environments used in parallel."""

    def __init__(
        self,
        environment_builder,
        worker_groups,
        workers_per_group,
        max_episode_steps,
    ):
        self.environment_builder = environment_builder
        self.worker_groups = worker_groups
        self.workers_per_group = workers_per_group
        self.max_episode_steps = max_episode_steps

    def initialize(self, seed):
        def proc(action_pipe, index, seed):
            """Process holding a sequential group of environments."""
            envs = Sequential(
                self.environment_builder,
                self.max_episode_steps,
                self.workers_per_group,
            )
            envs.initialize(seed)

            observations = envs.start()
            self.output_queue.put((index, observations))

            while True:
                actions = action_pipe.recv()
                out = envs.step(actions)
                self.output_queue.put((index, out))

        dummy_environment = self.environment_builder()
        self.observation_space = dummy_environment.observation_space
        self.action_space = dummy_environment.action_space
        del dummy_environment
        self.started = False

        self.output_queue = multiprocessing.Queue()
        self.action_pipes = []

        for i in range(self.worker_groups):
            pipe, worker_end = multiprocessing.Pipe()
            self.action_pipes.append(pipe)
            group_seed = seed + i * self.workers_per_group
            process = multiprocessing.Process(
                target=proc, args=(worker_end, i, group_seed)
            )
            process.daemon = True
            process.start()

    def start(self):
        """Used once to get the initial observations."""
        assert not self.started
        self.started = True
        observations_list = [None for _ in range(self.worker_groups)]

        for _ in range(self.worker_groups):
            index, observations = self.output_queue.get()
            observations_list[index] = observations

        self.observations_list = np.array(observations_list)
        self.next_observations_list = np.zeros_like(self.observations_list)
        self.rewards_list = np.zeros(
            (self.worker_groups, self.workers_per_group), np.float32
        )
        self.resets_list = np.zeros(
            (self.worker_groups, self.workers_per_group), bool
        )
        self.terminations_list = np.zeros(
            (self.worker_groups, self.workers_per_group), bool
        )

        return np.concatenate(self.observations_list)

    def step(self, actions):
        actions_list = np.split(actions, self.worker_groups)
        for actions, pipe in zip(actions_list, self.action_pipes):
            pipe.send(actions)

        for _ in range(self.worker_groups):
            index, (observations, infos) = self.output_queue.get()
            self.observations_list[index] = observations
            self.next_observations_list[index] = infos["observations"]
            self.rewards_list[index] = infos["rewards"]
            self.resets_list[index] = infos["resets"]
            self.terminations_list[index] = infos["terminations"]

        observations = np.concatenate(self.observations_list)
        infos = dict(
            observations=np.concatenate(self.next_observations_list),
            rewards=np.concatenate(self.rewards_list),
            resets=np.concatenate(self.resets_list),
            terminations=np.concatenate(self.terminations_list),
        )
        return observations, infos


def distribute(environment_builder, worker_groups=1, workers_per_group=1):
    """Distributes workers over parallel and sequential groups."""
    dummy_environment = environment_builder()
    max_episode_steps = dummy_environment.max_episode_steps
    del dummy_environment

    if worker_groups < 2:
        return Sequential(
            environment_builder,
            max_episode_steps=max_episode_steps,
            workers=workers_per_group,
        )

    return Parallel(
        environment_builder,
        worker_groups=worker_groups,
        workers_per_group=workers_per_group,
        max_episode_steps=max_episode_steps,
    )



from tonic import replays
from tonic.tensorflow import agents, models, normalizers, updaters


def default_model():
    return models.ActorTwinCriticWithTargets(
        actor=models.Actor(
            encoder=models.ObservationEncoder(),
            torso=models.MLP((256, 256), "relu"),
            head=models.DeterministicPolicyHead(),
        ),
        critic=models.Critic(
            encoder=models.ObservationActionEncoder(),
            torso=models.MLP((256, 256), "relu"),
            head=models.DistributionalValueHead(-150.0, 150.0, 51),
        ),
        observation_normalizer=normalizers.MeanStd(),
    )


class TD4(agents.TD3):
    def __init__(
        self,
        model=None,
        replay=None,
        exploration=None,
        actor_updater=None,
        critic_updater=None,
        delay_steps=2,
    ):
        model = model or default_model()
        replay = replay or replays.Buffer(return_steps=5)
        actor_updater = (
            actor_updater
            or updaters.DistributionalDeterministicPolicyGradient()
        )
        critic_updater = (
            critic_updater
            or updaters.TwinCriticDistributionalDeterministicQLearning()
        )
        super().__init__(
            model=model,
            replay=replay,
            exploration=exploration,
            actor_updater=actor_updater,
            critic_updater=critic_updater,
            delay_steps=delay_steps,
        )



import os
import random

import numpy as np
import torch

from deprl.vendor.tonic import agents
from deprl.vendor.tonic.utils import logger


class Agent(agents.Agent):
    def initialize(self, seed=None):
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)
            torch.manual_seed(seed)

    def save(self, path):
        path = path + ".pt"
        logger.log(f"\nSaving weights to {path}")
        os.makedirs(os.path.dirname(path), exist_ok=True)
        torch.save(self.model.state_dict(), path)

    def load(self, path):
        path = path + ".pt"
        logger.log(f"\nLoading weights from {path}")
        if torch.cuda.is_available():
            state_dict = torch.load(path)
        else:
            state_dict = torch.load(path, map_location="cpu")
        try:
            if "critic.torso.model.0.weight" in state_dict.keys():
                logger.log("Loading full model.")
                self._load_weights(state_dict, full=True)
            else:
                logger.log("Loading only actor weights.")
                self._load_weights(state_dict, full=False)
        except RuntimeError as e:
            logger.log(f"Loading failed, policy mismatch with checkpoint: {e}")

    def _load_weights(self, state_dict, full=False):
        if full:
            self.model.load_state_dict(state_dict)
        else:
            self.model.actor.load_state_dict(state_dict)



import tensorflow as tf
from tonic import logger
from tonic.tensorflow import agents, updaters


class TRPO(agents.A2C):
    """Trust Region Policy Optimization.
    TRPO: https://arxiv.org/pdf/1502.05477.pdf
    """

    def __init__(
        self, model=None, replay=None, actor_updater=None, critic_updater=None
    ):
        actor_updater = actor_updater or updaters.TrustRegionPolicyGradient()
        super().__init__(
            model=model,
            replay=replay,
            actor_updater=actor_updater,
            critic_updater=critic_updater,
        )

    def step(self, observations, steps):
        # Sample actions and get their log-probabilities for training.
        actions, log_probs, locs, scales = self._step(observations)
        actions = actions.numpy()
        log_probs = log_probs.numpy()
        locs = locs.numpy()
        scales = scales.numpy()

        # Keep some values for the next update.
        self.last_observations = observations.copy()
        self.last_actions = actions.copy()
        self.last_log_probs = log_probs.copy()
        self.last_locs = locs.copy()
        self.last_scales = scales.copy()

        return actions

    def update(self, observations, rewards, resets, terminations, steps):
        # Store the last transitions in the replay.
        self.replay.store(
            observations=self.last_observations,
            actions=self.last_actions,
            next_observations=observations,
            rewards=rewards,
            resets=resets,
            terminations=terminations,
            log_probs=self.last_log_probs,
            locs=self.last_locs,
            scales=self.last_scales,
        )

        # Prepare to update the normalizers.
        if self.model.observation_normalizer:
            self.model.observation_normalizer.record(self.last_observations)
        if self.model.return_normalizer:
            self.model.return_normalizer.record(rewards)

        # Update the model if the replay is ready.
        if self.replay.ready():
            self._update()

    @tf.function
    def _step(self, observations):
        distributions = self.model.actor(observations)
        if hasattr(distributions, "sample_with_log_prob"):
            actions, log_probs = distributions.sample_with_log_prob()
        else:
            actions = distributions.sample()
            log_probs = distributions.log_prob(actions)
        locs = distributions.loc
        scales = distributions.stddev()
        return actions, log_probs, locs, scales

    def _update(self):
        # Compute the lambda-returns.
        batch = self.replay.get_full("observations", "next_observations")
        values, next_values = self._evaluate(**batch)
        values, next_values = values.numpy(), next_values.numpy()
        self.replay.compute_returns(values, next_values)

        keys = (
            "observations",
            "actions",
            "log_probs",
            "locs",
            "scales",
            "advantages",
        )
        batch = self.replay.get_full(*keys)
        infos = self.actor_updater(**batch)
        for k, v in infos.items():
            logger.store("actor/" + k, v.numpy())

        critic_iterations = 0
        for batch in self.replay.get("observations", "returns"):
            infos = self.critic_updater(**batch)
            critic_iterations += 1
            for k, v in infos.items():
                logger.store("critic/" + k, v.numpy())
        logger.store("critic/iterations", critic_iterations)

        # Update the normalizers.
        if self.model.observation_normalizer:
            self.model.observation_normalizer.update()
        if self.model.return_normalizer:
            self.model.return_normalizer.update()



import os

import numpy as np
from matplotlib import pyplot as plt


# precision = 0.5 # slow but accurate
# precision = 0.5 # fast but inaccurate
# 0.1
precision = 0.1 # fast but inaccurate

FORCE_THRESHOLD = 0.01
GRF_LIST = ["leg1_r.grf_norm_y", "leg0_l.grf_norm_y"]
JOINT_LIST = [
    "hip_flexion_r",
    "knee_angle_r",
    "ankle_angle_r",
    "hip_flexion_l",
    "knee_angle_l",
    "ankle_angle_l",
    "time",
]


def deg_2_rad(x):
    return (x / 360) * (2 * np.pi)


def check_force(grf, time):
    if grf[time] > FORCE_THRESHOLD:
        return 1
    else:
        return 0


def extract_normalized(sto, begin, end, time):
    """
    Takes in a list of values <sto> and uses <time> to extend the duration using
    interpolation to a continuous percentage of the gait cycle. This is used
    to avoid small shifts in gait cycle timings from accumulating.
    """
    new_sto = []
    for perc in np.arange(0, 100, precision):
        t = begin + perc * (end - begin) / 100
        new_sto.append(interpolate(sto, time, t))
    return new_sto


def interpolate(sto, time, t):
    """
    Outputs the linear interpolation of <sto> at time <t>, where the discretized
    times are given by <time>.
    """
    lower_index = -1
    upper_index = -1
    for i in range(len(time)):
        if time[i] <= t:
            lower_index = i
        if time[i] > t:
            upper_index = i
            break
    if upper_index == -1:
        upper_index = lower_index
    if upper_index == lower_index and upper_index == len(time) - 1:
        return sto[-1]
    alpha = (t - time[lower_index]) / (time[upper_index] - time[lower_index])
    return sto[lower_index] * (1 - alpha) + sto[upper_index] * alpha


def get_gait_start_times(time, grf):
    """
    Find gait start times from foot contact set-downs.
    """
    grf = np.array(extract_normalized(grf, time[0], time[-1], time))
    time_extended = np.array(extract_normalized(time, time[0], time[-1], time))
    prev_force = 0
    gait_start_times = []
    for tdx in range(grf.shape[0]):
        force = check_force(grf, tdx)
        if not prev_force and force:
            gait_start_times.append(time_extended[tdx])
        prev_force = force
    return gait_start_times[1:]


# TODO find params that work for 2D gait perfectly, then try metrics on this. then either tune coeffs for 3D or tune mpo or dep for 3D.


def extract_gait_cycle(joint, time, grf):
    gait_start_times = get_gait_start_times(time, grf)
    gaits = []
    for idx in range(len(gait_start_times) - 1):
        gaits.append(
            extract_normalized(
                joint, gait_start_times[idx], gait_start_times[idx + 1], time
            )
        )
    return np.array(gaits)


def prepare_hip(arr):
    return arr


def prepare_ankle(arr):
    # offset in experimental data, read out from zml
    arr = [x + 0.349 for x in arr]
    return arr
    # return [x + 0.349 for x in arr]


def prepare_knee(arr):
    """
    prepare knee data. the mirroring around 0 switches the minimum and
    maximum arrays around, but tuples don't allow for item assignment.
    so we cast to list.
    """
    return arr


def normalize(arr, mini=None, maxi=None):
    if mini is None and maxi is None:
        mini = np.min(arr)
        maxi = np.max(arr)
    arr = np.array(arr)
    arr = (arr - np.min(arr)) / (np.max(arr) - np.min(arr))
    return np.asarray(arr), mini, maxi


def load_exp_data():
    """
    Save data as (mean, min, max) for each gait cycle.
    """
    file1 = open(os.path.join(os.path.dirname(__file__), "resources/exp_gait.zml"), "r")
    lines = file1.readlines()
    maxs = None
    means = None
    mins = None
    for line in lines:
        if "title" in line:
            title = " ".join(line.split()[-2:])
        if "norm_max" in line:
            maxs = line.split()[3:-1]
            maxs = np.array(maxs, dtype=np.float32)
        if "norm_min" in line:
            mins = line.split()[3:-1]
            mins = np.array(mins, dtype=np.float32)
        if "norm_mean" in line:
            means = line.split()[3:-1]
            means = np.array(means, dtype=np.float32)
        if maxs is not None and "Hip" in title:
            hip = (
                deg_2_rad(means.copy()),
                deg_2_rad(mins.copy()),
                deg_2_rad(maxs.copy()),
            )
            maxs = None
            mins = None
            means = None
        if maxs is not None and "Knee" in title:
            knee = (
                deg_2_rad(means.copy()),
                deg_2_rad(mins.copy()),
                deg_2_rad(maxs.copy()),
            )
            maxs = None
            mins = None
            means = None
        if maxs is not None and "Ankle" in title:
            ankle = (
                deg_2_rad(means.copy()),
                deg_2_rad(mins.copy()),
                deg_2_rad(maxs.copy()),
            )
            maxs = None
            means = None
            mins = None
    # return prepare_hip(hip), prepare_knee(knee), prepare_ankle(ankle)
    return prepare_hip(hip), prepare_knee(knee), prepare_ankle(ankle)


def euclid(query, template):
    length = np.minimum(len(query), len(template))
    return np.mean(np.square(np.array(query)[:length] - np.array(template)[:length]))


def get_avg_gait_cycle(joints, grf):
    gaits = []
    for gidx, jdxs in zip(range(2), [range(-6, -3), range(-3, 0)]):
        for joints_query, grf_query in zip(joints, grf):
            gaits_query = extract_gait_cycle(
                [x[jdxs] for x in joints_query],
                [x[0] for x in joints_query],
                remove_small_peaks([x[gidx] for x in grf_query]),
            )
            if gaits_query.shape[0] > 0:
                gaits.append(gaits_query)
            # gaits.append(gaits_template)
    gaits_mean = [np.mean(x, axis=0) for x in gaits]
    return gaits, gaits_mean


def filter_nans_out_of_list(arr):
    return arr[np.where(~np.isnan(np.array(arr)))]


def get_avg_dtwd(joints, grfs, plot=False, convert_fn=None):
    gaits, gaits_mean = get_avg_gait_cycle(joints, grfs)
    if convert_fn is not None:
        gaits, gaits_mean = convert_fn(gaits, gaits_mean)
    if len(gaits) == 0:
        return (0, 0, 0, 0)
    exp_data = load_exp_data()
    exp_data = [list(x) for x in exp_data]
    for i in range(len(exp_data)):
        for j in range(len(exp_data[i])):
            exp_data[i][j] = np.roll(exp_data[i][j], exp_data[i][j].shape[0] // 2, axis=0)
    if plot:
        fig, axs = plt.subplots(3, 1)
        fig, axs = get_plot(fig, axs, gaits, gaits_mean, color="grey")
        get_plot_exp(fig, axs, name="exp", color="grey", plot_mean=False)
        fig.savefig("test.png")
    time_exp = np.linspace(0, 1, 101)
    counter = [[], [], []]
    for gait in gaits_mean:
        time = np.linspace(0, 1, gait.shape[0])
        for g in range(3):
            for tidx, val in enumerate(gait[:, g]):
                if val < interpolate(
                    exp_data[g][2], time_exp, time[tidx]
                ) and val > interpolate(exp_data[g][1], time_exp, time[tidx]):
                    counter[g].append(1)
                else:
                    counter[g].append(0)
    return (
        np.mean(counter[0]) + 1 * np.mean(counter[1]) + 1 * np.mean(counter[2]),
        np.mean(counter[0]),
        np.mean(counter[1]),
        np.mean(counter[2]),
    )

def msd_get_avg_dtwd(joints, grfs, plot=False, convert_fn=None):
    gaits, gaits_mean = get_avg_gait_cycle(joints, grfs)
    exp_data = load_exp_data()
    for sto in gaits_mean:
        dtwds = np.zeros((3,))
        for g in range(3):
            interp_exp_data = []
            for t in np.linspace(0, 1, sto.shape[0]):
                interp_exp_data.append(interpolate(exp_data[g][0], np.linspace(0, 1, 101), t))
            dtwds[g] = np.mean(np.square(interp_exp_data - sto[:, g]))
    return np.sum(dtwds), dtwds[0], dtwds[1], dtwds[2]

def new_get_avg_dtwd(joints, grfs, plot=False, convert_fn=None):
    gaits, gaits_mean = get_avg_gait_cycle(joints, grfs)
    if convert_fn is not None:
        gaits, gaits_mean = convert_fn(gaits, gaits_mean)
    exp_data = load_exp_data()
    exp_data = [list(x) for x in exp_data]
    dtwd_collector = []
    for sto in gaits_mean:
        dtwds = np.zeros((3,))
        for g in range(3):
            interp_exp_data = []
            for t in np.linspace(0, 1, sto.shape[0]):
                interp_exp_data.append([interpolate(exp_data[g][j], np.linspace(0, 1, 101), t) for j in range(3)])
            interp_exp_data = np.array(interp_exp_data)
            arr1 = np.where(interp_exp_data[:, 1] < sto[:, g])[0]
            arr2 = np.where(interp_exp_data[:, 2] > sto[:, g])[0]
            common = np.intersect1d(arr1, arr2)
            dtwds[g] = common.shape[0] / interp_exp_data[:, 0].shape[0]
        dtwd_collector.append(dtwds)
    if plot:
        fignew, axnew = plt.subplots(3, 1)
        axnew[0].plot(np.linspace(0, 1, exp_data[0][0].shape[0]), exp_data[0][0], color='blue')
        axnew[0].plot(np.linspace(0, 1, exp_data[0][1].shape[0]), exp_data[0][1], color='red')
        axnew[0].plot(np.linspace(0, 1, exp_data[0][2].shape[0]), exp_data[0][2], color='green')
        axnew[0].plot(np.linspace(0, 1, gaits_mean[0][:,0].shape[0]), gaits_mean[0][:, 0], color='black')
        plt.savefig('testimage.pdf')
    dtwd_mean = [np.mean([x[i] for x in dtwd_collector]) for i in range(3)]
    return np.sum(dtwd_mean), dtwd_mean[0], dtwd_mean[1], dtwd_mean[2]




def get_plot(fig, axs, gaits, gaits_mean, color="tab:red", name=None, index=0):
    names = ["hip", "knee", "ankle"]
    plot_all = False
    plot_mean = True
    if plot_all:
        for sto in gaits:
            for gait in sto:
                for g in range(3):
                    axs[g].plot(
                        np.linspace(0, 1, gait.shape[0]),
                        gait[:, g],
                        color=color,
                        alpha=0.1,
                    )
                    axs[g, index].set_xlim([0, 1])
                    axs[g, 0].set_ylabel(f"{names[g]} [rad]")
    # if plot_mean:
    #     for gait in gaits_mean:
    #         for g in range(3):
    #             axs[g, index].plot(
    #                 np.linspace(0, 1, gait.shape[0]), gait[:, g], color=color, alpha=0.2
    #             )
    #             axs[g, index].set_xlim([0, 1])
    #             axs[g, 0].set_ylabel(f"{names[g]} [rad]")
    hip = np.mean([x[:, 0] for x in gaits_mean], axis=0)
    knee = np.mean([x[:, 1] for x in gaits_mean], axis=0)
    ankle = np.mean([x[:, 2] for x in gaits_mean], axis=0)
    axs[0, index].plot(np.linspace(0, 1, hip.shape[0]), hip, color="tab:red", linewidth=1)
    axs[1, index].plot(np.linspace(0, 1, knee.shape[0]), knee, color="tab:red", linewidth=1)
    axs[2, index].plot(np.linspace(0, 1, ankle.shape[0]), ankle, color="tab:red", linewidth=1)
    if name is not None:
        plt.savefig(f"{name}.pdf")
        return
    return fig, axs


def get_plot_exp(fig, axs, color="tab:red", name="none", plot_mean=True, index=0):
    exp_data = load_exp_data()
    exp_data = [list(x) for x in exp_data]
    # for i in range(len(joints)):
    #     for j in range(len(joints[i])):
    #         joints[i][j] = np.roll(joints[i][j], joints[i][j].shape[0] // 2, axis=0)
    hip, knee, ankle = exp_data
    if plot_mean:
        axs[0, index].plot(
            np.linspace(0, 1, hip[0].shape[0]), hip[0], color=color, linewidth=3
        )
    axs[0, index].fill_between(
        np.linspace(0, 1, hip[0].shape[0]),
        hip[1],
        hip[2],
        color=color,
        linewidth=3,
        alpha=0.2,
    )
    if plot_mean:
        axs[1, index].plot(
            np.linspace(0, 1, knee[0].shape[0]), knee[0], color=color, linewidth=3
        )
    axs[1, index].fill_between(
        np.linspace(0, 1, knee[0].shape[0]),
        knee[1],
        knee[2],
        color=color,
        linewidth=3,
        alpha=0.2,
    )
    if plot_mean:
        axs[2, index].plot(
            np.linspace(0, 1, ankle[0].shape[0]), ankle[0], color=color, linewidth=3
        )
    axs[2, index].fill_between(
        np.linspace(0, 1, ankle[0].shape[0]),
        ankle[1],
        ankle[2],
        color=color,
        linewidth=3,
        alpha=0.2,
    )
    axs[2, index].set_xlabel("gait cycle [%]")
    names = ["hip", "knee", "ankle"]
    for i in range(axs.shape[0]):
        axs[i, 0].set_ylabel(f"{names[i]} [rad]")
        axs[i, index].set_xlim([0, 1])
    return fig, axs


def list_of_strings_2_string(list_of_strings):
    string = ""
    for s in list_of_strings:
        string += s
    return string


"""
for sto in sto_files:
    path = os.path.join(folder, sto)
    joints_rl, grf_rl = read_joints(name=path, color='tab:red')
    joints_geyer, grf_geyer = read_joints(name='geyer.sto', color='tab:blue')

    ret = 0
    for gidx, jdxs in zip(range(2), [[-6, -5, -4], [-3, -2, -1]]):
        ret += np.mean([get_sample_dtwd(joints_rl, grf_rl[:, gidx], joints_geyer, grf_geyer[:, gidx], jdx) for jdx in jdxs])
    return [ret if not np.isnan(ret) else 200][0]
"""


def readMotionFile(filename):
    """Reads OpenSim .sto files.
    Parameters
    ----------
    filename: absolute path to the .sto file
    Returns
    -------
    header: the header of the .sto
    labels: the labels of the columns
    data: an array of the data
    """
    filename = os.path.join(os.path.abspath(__file__), "resources/", filename)
    if not os.path.exists(filename):
        print("file do not exists")

    file_id = open(filename, "r")

    # read header
    next_line = file_id.readline()
    header = [next_line]
    nc = 0
    nr = 0
    while "endheader" not in next_line:
        if "datacolumns" in next_line:
            nc = int(next_line[next_line.index(" ") + 1 : len(next_line)])
        elif "datarows" in next_line:
            nr = int(next_line[next_line.index(" ") + 1 : len(next_line)])
        elif "nColumns" in next_line:
            nc = int(next_line[next_line.index("=") + 1 : len(next_line)])
        elif "nRows" in next_line:
            nr = int(next_line[next_line.index("=") + 1 : len(next_line)])

        next_line = file_id.readline()
        header.append(next_line)

    # process column labels
    next_line = file_id.readline()
    if next_line.isspace() is True:
        next_line = file_id.readline()

    labels = next_line.split()

    # get data
    data = []
    for i in range(1, nr + 1):
        d = [float(x) for x in file_id.readline().split()]
        data.append(d)

    file_id.close()

    return header, labels, data, nc, nr


def read_joints(name):
    ret = readMotionFile(name)
    joints = []
    grf = []
    for time in ret[2]:
        joint_step = []
        grf_step = []
        for idx, element in enumerate(time):
            if ret[1][idx] in JOINT_LIST:
                joint_step.append(element)
            if ret[1][idx] in GRF_LIST:
                grf_step.append(element)
        joints.append(joint_step)
        grf.append(grf_step)
    joints = np.array(joints)
    grf = np.array(grf)
    return joints, grf

def read_energy(name):
    ret = readMotionFile(name)
    acts = []
    for time in ret[2]:
        act_step = []
        for idx, element in enumerate(time):
            if "activation" in ret[1][idx]:
                act_step.append(element)
        acts.append(act_step)
    acts = np.array(acts)
    return np.mean(np.power(acts, 3))


def remove_small_peaks(arr, sensitivity=FORCE_THRESHOLD, threshold=0.8):
    arr = np.asarray(arr)
    peak_on, prev_peak, pmax = 0, 0, 0
    for i in range(arr.shape[0]):
        peak_on = [1 if arr[i] > sensitivity else 0][0]
        if peak_on and not prev_peak:
            start = i
        if peak_on and prev_peak:
            pmax = max(pmax, arr[i])
        if not peak_on and prev_peak:
            if pmax < threshold:
                arr[start:i] = 0
            pmax = 0
        prev_peak = peak_on
    return arr



from . import custom_agents, custom_mpo_torch, custom_trainer
from .env_wrappers import apply_wrapper, env_tonic_compat
from .utils import load, load_baseline, mujoco_render
from .vendor.tonic import (
    Trainer,
    agents,
    environments,
    explorations,
    logger,
    replays,
    torch,
)

__all__ = [
    custom_mpo_torch,
    custom_agents,
    custom_trainer,
    apply_wrapper,
    env_tonic_compat,
    torch,
    agents,
    environments,
    explorations,
    logger,
    replays,
    Trainer,
    load,
    load_baseline,
    mujoco_render,
]



import os
import time

import numpy as np

from deprl.vendor.tonic.utils import logger


class Trainer:
    """Trainer used to train and evaluate an agent on an environment."""

    def __init__(
        self,
        steps=int(1e7),
        epoch_steps=int(2e4),
        save_steps=int(5e5),
        test_episodes=5,
        show_progress=True,
        replace_checkpoint=False,
    ):
        self.max_steps = steps
        self.epoch_steps = epoch_steps
        self.save_steps = save_steps
        self.test_episodes = test_episodes
        self.show_progress = show_progress
        self.replace_checkpoint = replace_checkpoint

    def initialize(self, agent, environment, test_environment=None):
        self.agent = agent
        self.environment = environment
        self.test_environment = test_environment

    def run(self):
        """Runs the main training loop."""

        start_time = last_epoch_time = time.time()

        # Start the environments.
        observations = self.environment.start()

        num_workers = len(observations)
        scores = np.zeros(num_workers)
        lengths = np.zeros(num_workers, int)
        self.steps, epoch_steps, epochs, episodes = 0, 0, 0, 0
        steps_since_save = 0

        while True:
            # Select actions.
            actions = self.agent.step(observations, self.steps)
            assert not np.isnan(actions.sum())
            logger.store("train/action", actions, stats=True)

            # Take a step in the environments.
            observations, infos = self.environment.step(actions)
            self.agent.update(**infos, steps=self.steps)

            scores += infos["rewards"]
            lengths += 1
            self.steps += num_workers
            epoch_steps += num_workers
            steps_since_save += num_workers

            # Show the progress bar.
            if self.show_progress:
                logger.show_progress(
                    self.steps, self.epoch_steps, self.max_steps
                )

            # Check the finished episodes.
            for i in range(num_workers):
                if infos["resets"][i]:
                    logger.store("train/episode_score", scores[i], stats=True)
                    logger.store(
                        "train/episode_length", lengths[i], stats=True
                    )
                    scores[i] = 0
                    lengths[i] = 0
                    episodes += 1

            # End of the epoch.
            if epoch_steps >= self.epoch_steps:
                # Evaluate the agent on the test environment.
                if self.test_environment:
                    self._test()

                # Log the data.
                epochs += 1
                current_time = time.time()
                epoch_time = current_time - last_epoch_time
                sps = epoch_steps / epoch_time
                logger.store("train/episodes", episodes)
                logger.store("train/epochs", epochs)
                logger.store("train/seconds", current_time - start_time)
                logger.store("train/epoch_seconds", epoch_time)
                logger.store("train/epoch_steps", epoch_steps)
                logger.store("train/steps", self.steps)
                logger.store("train/worker_steps", self.steps // num_workers)
                logger.store("train/steps_per_second", sps)
                logger.dump()
                last_epoch_time = time.time()
                epoch_steps = 0

            # End of training.
            stop_training = self.steps >= self.max_steps

            # Save a checkpoint.
            if stop_training or steps_since_save >= self.save_steps:
                path = os.path.join(logger.get_path(), "checkpoints")
                if os.path.isdir(path) and self.replace_checkpoint:
                    for file in os.listdir(path):
                        if file.startswith("step_"):
                            os.remove(os.path.join(path, file))
                checkpoint_name = f"step_{self.steps}"
                save_path = os.path.join(path, checkpoint_name)
                self.agent.save(save_path)
                steps_since_save = self.steps % self.save_steps

            if stop_training:
                break

    def _test(self):
        """Tests the agent on the test environment."""

        # Start the environment.
        if not hasattr(self, "test_observations"):
            self.test_observations = self.test_environment.start()
            assert len(self.test_observations) == 1

        # Test loop.
        for _ in range(self.test_episodes):
            score, length = 0, 0

            while True:
                # Select an action.
                actions = self.agent.test_step(
                    self.test_observations, self.steps
                )
                assert not np.isnan(actions.sum())
                logger.store("test/action", actions, stats=True)

                # Take a step in the environment.
                self.test_observations, infos = self.test_environment.step(
                    actions
                )
                self.agent.test_update(**infos, steps=self.steps)

                score += infos["rewards"][0]
                length += 1

                if infos["resets"][0]:
                    break

            # Log the data.
            logger.store("test/episode_score", score, stats=True)
            logger.store("test/episode_length", length, stats=True)



import torch

from deprl.vendor.tonic.torch import models, updaters  # noqa


class VRegression:
    def __init__(self, loss=None, optimizer=None, gradient_clip=0):
        self.loss = loss or torch.nn.MSELoss()
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-3)
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.critic)
        self.optimizer = self.optimizer(self.variables)

    def __call__(self, observations, returns):
        self.optimizer.zero_grad()
        values = self.model.critic(observations)
        loss = self.loss(values, returns)

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        return dict(loss=loss.detach(), v=values.detach())


class QRegression:
    def __init__(self, loss=None, optimizer=None, gradient_clip=0):
        self.loss = loss or torch.nn.MSELoss()
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-3)
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.critic)
        self.optimizer = self.optimizer(self.variables)

    def __call__(self, observations, actions, returns):
        self.optimizer.zero_grad()
        values = self.model.critic(observations, actions)
        loss = self.loss(values, returns)

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        return dict(loss=loss.detach(), q=values.detach())


class DeterministicQLearning:
    def __init__(self, loss=None, optimizer=None, gradient_clip=0):
        self.loss = loss or torch.nn.MSELoss()
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-3)
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.critic)
        self.optimizer = self.optimizer(self.variables)

    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        with torch.no_grad():
            next_actions = self.model.target_actor(next_observations)
            next_values = self.model.target_critic(
                next_observations, next_actions
            )
            returns = rewards + discounts * next_values

        self.optimizer.zero_grad()
        values = self.model.critic(observations, actions)
        loss = self.loss(values, returns)

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        return dict(loss=loss.detach(), q=values.detach())


class DistributionalDeterministicQLearning:
    def __init__(self, optimizer=None, gradient_clip=0):
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-3)
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.critic)
        self.optimizer = self.optimizer(self.variables)

    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        with torch.no_grad():
            next_actions = self.model.target_actor(next_observations)
            next_value_distributions = self.model.target_critic(
                next_observations, next_actions
            )
            values = next_value_distributions.values
            returns = rewards[:, None] + discounts[:, None] * values
            targets = next_value_distributions.project(returns)

        self.optimizer.zero_grad()
        value_distributions = self.model.critic(observations, actions)
        log_probabilities = torch.nn.functional.log_softmax(
            value_distributions.logits, dim=-1
        )
        loss = -(targets * log_probabilities).sum(dim=-1).mean()

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        return dict(loss=loss.detach())


class TargetActionNoise:
    def __init__(self, scale=0.2, clip=0.5):
        self.scale = scale
        self.clip = clip

    def __call__(self, actions):
        noises = self.scale * torch.randn_like(actions)
        noises = torch.clamp(noises, -self.clip, self.clip)
        actions = actions + noises
        return torch.clamp(actions, -1, 1)


class TwinCriticDeterministicQLearning:
    def __init__(
        self,
        loss=None,
        optimizer=None,
        target_action_noise=None,
        gradient_clip=0,
    ):
        self.loss = loss or torch.nn.MSELoss()
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-3)
        )
        self.target_action_noise = target_action_noise or TargetActionNoise(
            scale=0.2, clip=0.5
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        variables_1 = models.trainable_variables(self.model.critic_1)
        variables_2 = models.trainable_variables(self.model.critic_2)
        self.variables = variables_1 + variables_2
        self.optimizer = self.optimizer(self.variables)

    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        with torch.no_grad():
            next_actions = self.model.target_actor(next_observations)
            next_actions = self.target_action_noise(next_actions)
            next_values_1 = self.model.target_critic_1(
                next_observations, next_actions
            )
            next_values_2 = self.model.target_critic_2(
                next_observations, next_actions
            )
            next_values = torch.min(next_values_1, next_values_2)
            returns = rewards + discounts * next_values

        self.optimizer.zero_grad()
        values_1 = self.model.critic_1(observations, actions)
        values_2 = self.model.critic_2(observations, actions)
        loss_1 = self.loss(values_1, returns)
        loss_2 = self.loss(values_2, returns)
        loss = loss_1 + loss_2

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        return dict(
            loss=loss.detach(), q1=values_1.detach(), q2=values_2.detach()
        )


class TwinCriticSoftQLearning:
    def __init__(
        self, loss=None, optimizer=None, entropy_coeff=0.2, gradient_clip=0
    ):
        self.loss = loss or torch.nn.MSELoss()
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=3e-4)
        )
        self.entropy_coeff = entropy_coeff
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        variables_1 = models.trainable_variables(self.model.critic_1)
        variables_2 = models.trainable_variables(self.model.critic_2)
        self.variables = variables_1 + variables_2
        self.optimizer = self.optimizer(self.variables)

    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        with torch.no_grad():
            next_distributions = self.model.actor(next_observations)
            if hasattr(next_distributions, "rsample_with_log_prob"):
                outs = next_distributions.rsample_with_log_prob()
                next_actions, next_log_probs = outs
            else:
                next_actions = next_distributions.rsample()
                next_log_probs = next_distributions.log_prob(next_actions)
            next_log_probs = next_log_probs.sum(dim=-1)
            next_values_1 = self.model.target_critic_1(
                next_observations, next_actions
            )
            next_values_2 = self.model.target_critic_2(
                next_observations, next_actions
            )
            next_values = torch.min(next_values_1, next_values_2)
            returns = rewards + discounts * (
                next_values - self.entropy_coeff * next_log_probs
            )

        self.optimizer.zero_grad()
        values_1 = self.model.critic_1(observations, actions)
        values_2 = self.model.critic_2(observations, actions)
        loss_1 = self.loss(values_1, returns)
        loss_2 = self.loss(values_2, returns)
        loss = loss_1 + loss_2

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        return dict(
            loss=loss.detach(), q1=values_1.detach(), q2=values_2.detach()
        )


class ExpectedSARSA:
    def __init__(
        self, num_samples=20, loss=None, optimizer=None, gradient_clip=0
    ):
        self.num_samples = num_samples
        self.loss = loss or torch.nn.MSELoss()
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=3e-4)
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.critic)
        self.optimizer = self.optimizer(self.variables)

    def __call__(
        self, observations, actions, next_observations, rewards, discounts
    ):
        # Approximate the expected next values.
        with torch.no_grad():
            next_target_distributions = self.model.target_actor(
                next_observations
            )
            next_actions = next_target_distributions.rsample(
                (self.num_samples,)
            )
            next_actions = updaters.merge_first_two_dims(next_actions)
            next_observations = updaters.tile(
                next_observations, self.num_samples
            )
            next_observations = updaters.merge_first_two_dims(
                next_observations
            )
            next_values = self.model.target_critic(
                next_observations, next_actions
            )
            next_values = next_values.view(self.num_samples, -1)
            next_values = next_values.mean(dim=0)
            returns = rewards + discounts * next_values

        self.optimizer.zero_grad()
        values = self.model.critic(observations, actions)
        loss = self.loss(returns, values)

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        return dict(loss=loss.detach(), q=values.detach())



from . import agents, environments, explorations, replays, torch
from .utils import logger
from .utils.trainer import Trainer

__all__ = [torch, agents, environments, explorations, logger, replays, Trainer]



import tensorflow as tf


def default_dense_kwargs():
    return dict(
        kernel_initializer=tf.keras.initializers.VarianceScaling(
            scale=1 / 3, mode="fan_in", distribution="uniform"
        ),
        bias_initializer=tf.keras.initializers.VarianceScaling(
            scale=1 / 3, mode="fan_in", distribution="uniform"
        ),
    )


def mlp(units, activation, dense_kwargs=None):
    if dense_kwargs is None:
        dense_kwargs = default_dense_kwargs()
    layers = [
        tf.keras.layers.Dense(u, activation, **dense_kwargs) for u in units
    ]
    return tf.keras.Sequential(layers)


MLP = mlp



import random

import numpy as np
import tensorflow as tf
from tonic import agents, logger


class Agent(agents.Agent):
    def initialize(self, seed=None):
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)
            tf.random.set_seed(seed)

    def save(self, path):
        logger.log(f"\nSaving weights to {path}")
        self.model.save_weights(path)

    def load(self, path):
        logger.log(f"\nLoading weights from {path}")
        self.model.load_weights(path)



import tensorflow as tf
from tonic import explorations
from tonic.tensorflow import agents, models, normalizers, updaters


def default_model():
    return models.ActorTwinCriticWithTargets(
        actor=models.Actor(
            encoder=models.ObservationEncoder(),
            torso=models.MLP((256, 256), "relu"),
            head=models.GaussianPolicyHead(
                loc_activation=None,
                distribution=models.SquashedMultivariateNormalDiag,
            ),
        ),
        critic=models.Critic(
            encoder=models.ObservationActionEncoder(),
            torso=models.MLP((256, 256), "relu"),
            head=models.ValueHead(),
        ),
        observation_normalizer=normalizers.MeanStd(),
    )


class SAC(agents.DDPG):
    """Soft Actor-Critic.
    SAC: https://arxiv.org/pdf/1801.01290.pdf
    """

    def __init__(
        self,
        model=None,
        replay=None,
        exploration=None,
        actor_updater=None,
        critic_updater=None,
    ):
        model = model or default_model()
        exploration = exploration or explorations.NoActionNoise()
        actor_updater = (
            actor_updater
            or updaters.TwinCriticSoftDeterministicPolicyGradient()
        )
        critic_updater = critic_updater or updaters.TwinCriticSoftQLearning()
        super().__init__(
            model=model,
            replay=replay,
            exploration=exploration,
            actor_updater=actor_updater,
            critic_updater=critic_updater,
        )

    @tf.function
    def _stochastic_actions(self, observations):
        return self.model.actor(observations).sample()

    def _policy(self, observations):
        return self._stochastic_actions(observations).numpy()

    @tf.function
    def _greedy_actions(self, observations):
        return self.model.actor(observations).mode()



from .actors import ClippedRatio  # noqa
from .actors import (
    DeterministicPolicyGradient,
    DistributionalDeterministicPolicyGradient,
    MaximumAPosterioriPolicyOptimization,
    StochasticPolicyGradient,
    TrustRegionPolicyGradient,
    TwinCriticSoftDeterministicPolicyGradient,
)
from .critics import (
    DeterministicQLearning,
    DistributionalDeterministicQLearning,
    ExpectedSARSA,
    QRegression,
    TargetActionNoise,
    TwinCriticDeterministicQLearning,
    TwinCriticSoftQLearning,
    VRegression,
)
from .optimizers import ConjugateGradient
from .utils import merge_first_two_dims, tile

__all__ = [
    merge_first_two_dims,
    tile,
    ClippedRatio,
    DeterministicPolicyGradient,
    DistributionalDeterministicPolicyGradient,
    MaximumAPosterioriPolicyOptimization,
    StochasticPolicyGradient,
    TrustRegionPolicyGradient,
    TwinCriticSoftDeterministicPolicyGradient,
    DeterministicQLearning,
    DistributionalDeterministicQLearning,
    ExpectedSARSA,
    QRegression,
    TargetActionNoise,
    TwinCriticDeterministicQLearning,
    TwinCriticSoftQLearning,
    VRegression,
    ConjugateGradient,
]



import torch


class MLP(torch.nn.Module):
    def __init__(self, sizes, activation, fn=None):
        super().__init__()
        self.sizes = sizes
        self.activation = activation
        self.fn = fn

    def initialize(self, input_size):
        sizes = [input_size] + list(self.sizes)
        layers = []
        for i in range(len(sizes) - 1):
            layers += [
                torch.nn.Linear(sizes[i], sizes[i + 1]),
                self.activation(),
            ]
        self.model = torch.nn.Sequential(*layers)
        if self.fn is not None:
            self.model.apply(self.fn)
        return sizes[-1]

    def forward(self, inputs):
        return self.model(inputs)


def trainable_variables(model):
    return [p for p in model.parameters() if p.requires_grad]



import tensorflow as tf
from tonic import explorations, logger, replays
from tonic.tensorflow import agents, models, normalizers, updaters


def default_model():
    return models.ActorCriticWithTargets(
        actor=models.Actor(
            encoder=models.ObservationEncoder(),
            torso=models.MLP((256, 256), "relu"),
            head=models.DeterministicPolicyHead(),
        ),
        critic=models.Critic(
            encoder=models.ObservationActionEncoder(),
            torso=models.MLP((256, 256), "relu"),
            head=models.ValueHead(),
        ),
        observation_normalizer=normalizers.MeanStd(),
    )


class DDPG(agents.Agent):
    """Deep Deterministic Policy Gradient.
    DDPG: https://arxiv.org/pdf/1509.02971.pdf
    """

    def __init__(
        self,
        model=None,
        replay=None,
        exploration=None,
        actor_updater=None,
        critic_updater=None,
    ):
        self.model = model or default_model()
        self.replay = replay or replays.Buffer()
        self.exploration = exploration or explorations.NormalActionNoise()
        self.actor_updater = (
            actor_updater or updaters.DeterministicPolicyGradient()
        )
        self.critic_updater = (
            critic_updater or updaters.DeterministicQLearning()
        )

    def initialize(self, observation_space, action_space, seed=None):
        super().initialize(seed=seed)
        self.model.initialize(observation_space, action_space)
        self.replay.initialize(seed)
        self.exploration.initialize(self._policy, action_space, seed)
        self.actor_updater.initialize(self.model)
        self.critic_updater.initialize(self.model)

    def step(self, observations, steps):
        # Get actions from the actor and exploration method.
        actions = self.exploration(observations, steps)

        # Keep some values for the next update.
        self.last_observations = observations.copy()
        self.last_actions = actions.copy()

        return actions

    def test_step(self, observations, steps):
        # Greedy actions for testing.
        return self._greedy_actions(observations).numpy()

    def update(self, observations, rewards, resets, terminations, steps):
        # Store the last transitions in the replay.
        self.replay.store(
            observations=self.last_observations,
            actions=self.last_actions,
            next_observations=observations,
            rewards=rewards,
            resets=resets,
            terminations=terminations,
        )

        # Prepare to update the normalizers.
        if self.model.observation_normalizer:
            self.model.observation_normalizer.record(self.last_observations)
        if self.model.return_normalizer:
            self.model.return_normalizer.record(rewards)

        # Update the model if the replay is ready.
        if self.replay.ready(steps):
            self._update(steps)

        self.exploration.update(resets)

    @tf.function
    def _greedy_actions(self, observations):
        return self.model.actor(observations)

    def _policy(self, observations):
        return self._greedy_actions(observations).numpy()

    def _update(self, steps):
        keys = (
            "observations",
            "actions",
            "next_observations",
            "rewards",
            "discounts",
        )

        # Update both the actor and the critic multiple times.
        for batch in self.replay.get(*keys, steps=steps):
            infos = self._update_actor_critic(**batch)

            for key in infos:
                for k, v in infos[key].items():
                    logger.store(key + "/" + k, v.numpy())

        # Update the normalizers.
        if self.model.observation_normalizer:
            self.model.observation_normalizer.update()
        if self.model.return_normalizer:
            self.model.return_normalizer.update()

    @tf.function
    def _update_actor_critic(
        self, observations, actions, next_observations, rewards, discounts
    ):
        critic_infos = self.critic_updater(
            observations, actions, next_observations, rewards, discounts
        )
        actor_infos = self.actor_updater(observations)
        self.model.update_targets()
        return dict(critic=critic_infos, actor=actor_infos)
